"""

Dataset sub-module for fusion-tools

"""


import os
import sys
import random
import json
from math import floor

import large_image.constants
import numpy as np
import pandas as pd
import geopandas as gpd
import large_image
from tqdm import tqdm

from fusion_tools.utils.shapes import process_filters_queries

from shapely.geometry import shape, box, Polygon
from shapely.validation import make_valid
from shapely.ops import unary_union
from skimage.filters import threshold_otsu
from skimage.morphology import remove_small_holes
from skimage.measure import label, find_contours
from skimage.draw import polygon2mask
from PIL import Image
from io import BytesIO

from typing import Callable
from typing_extensions import Union

from joblib import Parallel, delayed


class FUSIONDataset:
    def __init__(self,
                 slides: list = [],
                 annotations: Union[list,dict,None] = None,
                 image_transforms: Union[Callable, None] = None,
                 target_transforms: Union[Callable, None] = None,
                 patch_size: list = [256,256],
                 patch_mode: str = 'bbox',
                 patch_region: Union[str,list,dict] = 'all',
                 overlap: Union[float,None] = None,
                 spatial_filters: Union[list,None] = None,
                 property_filters: Union[list,None] = None,
                 use_cache: bool = False,
                 shuffle: bool = True,
                 seed_val: int = 1701,
                 verbose: bool = False,
                 use_parallel: bool = True,
                 n_jobs: int = 1):
        
        self.slides = slides
        self.pre_annotations = annotations
        self.image_transforms = image_transforms
        self.target_transforms = target_transforms
        self.patch_size = patch_size
        self.patch_mode = patch_mode
        self.patch_region = patch_region
        self.overlap = overlap
        self.spatial_filters = spatial_filters
        self.property_filters = property_filters
        self.use_cache = use_cache
        self.shuffle = shuffle
        self.seed_val = seed_val
        self.verbose = verbose
        self.use_parallel = use_parallel
        self.n_jobs = n_jobs

        # Checking inputs
        assert self.patch_mode in ['all','even','overlap','bbox','centered_bbox','random_bbox']
        if self.pre_annotations is None:
            if 'bbox' in self.patch_mode:
                raise AssertionError('If no annotations are passed, patch_mode must be "all"')
        if type(self.patch_region)==str:
            assert self.patch_region in ['all','tissue']
        elif type(self.patch_region)==list:
            assert all([any([i==j for j in ['all','tissue']]) if type(i)==str else type(i)==dict for i in self.patch_region])
        
        # Setting seed for reproducibility (🖖)
        self.seed_val = seed_val
        self.plant_seed(seed_val)
        self.verbose = verbose
        self.use_parallel = use_parallel
        self.n_jobs = n_jobs

    def plant_seed(self, seed_val):

        random.seed(seed_val)
        os.environ["PYTHONHASHSEED"] = str(seed_val)
        np.random.seed(seed_val)
        #torch.manual_seed(seed_val)
        #torch.cuda.manual_seed(seed_val)

    def process_annotations(self):
        pass

    def gen_slide_patches(self):
        pass

    def find_tissue(self):
        pass

    def __len__(self):
        pass

    def __getitem__(self, idx:int):
        pass

    def __str__(self):
        pass




class SegmentationDataset:
    def __init__(self,
                 slides: list = [],
                 annotations: Union[list,dict,None] = None,
                 transforms: Union[None, Callable] = None,
                 target_transforms: Union[None, Callable] = None,
                 mask_key: Union[dict,None] = None,
                 patch_size: list = [256,256],
                 patch_mode: str = 'bbox',
                 patch_region: Union[str,list,dict] = 'all',
                 spatial_filters: Union[list,None] = None,
                 property_filters: Union[list,None] = None,
                 use_structures: Union[str,list] = 'all',
                 use_cache: bool = True,
                 shuffle: bool = True,
                 seed_val: int = 1701,
                 verbose: bool = False,
                 use_parallel: bool = True,
                 n_jobs: int = 1
                 ):
        
        self.slides = slides
        self.pre_annotations = annotations
        self.transforms = transforms
        self.target_transforms = target_transforms
        self.mask_key = mask_key
        self.patch_size = patch_size
        self.patch_mode = patch_mode
        self.patch_region = patch_region
        assert self.patch_mode in ['all','even','overlap','bbox','centered_bbox','random_bbox']
        if self.pre_annotations is None:
            if 'bbox' in self.patch_mode:
                raise AssertionError('If no annotations are passed, patch_mode must be "all"')
        if type(self.patch_region)==str:
            assert self.patch_region in ['all','tissue']
        elif type(self.patch_region)==list:
            assert all([any([i==j for j in ['all','tissue']]) if type(i)==str else type(i)==dict for i in self.patch_region])

        # These two (if not None) should both be lists of either dicts containing filters to apply
        # to all slides or lists of lists of filters to apply to each slide, can also be mixed with some 
        # filters applied to some slides and then None applied to others 
        # (see fusion_tools.shapes.process_filters_queries for specifications)
        self.spatial_filters = spatial_filters if not spatial_filters is None else []
        self.property_filters = property_filters if not property_filters is None else []
        
        self.use_structures = use_structures
        self.use_cache = use_cache
        self.shuffle = shuffle

        # Setting seed for reproducibility (🖖)
        self.seed_val = seed_val
        self.plant_seed(seed_val)
        self.verbose = verbose
        self.use_parallel = use_parallel
        self.n_jobs = n_jobs

        self.process_annotations()
        self.gen_slide_patches()

    def plant_seed(self, seed_val):

        random.seed(seed_val)
        os.environ["PYTHONHASHSEED"] = str(seed_val)
        np.random.seed(seed_val)
        #torch.manual_seed(seed_val)
        #torch.cuda.manual_seed(seed_val)
        
    def process_annotations(self):

        # Starting out with assuming all GeoJSON type
        all_names = []
        self.slide_annotation_names = []
        if not self.pre_annotations is None:
            if type(self.pre_annotations)==dict:
                self.annotations = [gpd.GeoDataFrame.from_features(self.pre_annotations['features'])]
                all_names.append(self.pre_annotations['properties']['name'])
                self.slide_annotation_names.extend(all_names)
            else:
                self.annotations = []
                for a in self.pre_annotations:
                    if type(a)==dict:
                        self.annotations.append(gpd.GeoDataFrame.from_features(a['features']))
                        all_names.append(a['properties']['name'])
                        self.slide_annotation_names.append([a['properties']['name']])
                    elif type(a)==list:
                        self.annotations.append([gpd.GeoDataFrame.from_features(b['features']) for b in a if type(b)==dict])
                        all_names.extend([b['properties']['name'] for b in a if type(b)==dict])        
                        self.slide_annotation_names.append([b['properties']['name'] for b in a if type(b)==dict])
            
            if self.mask_key is None:
                self.mask_key = {
                    name:idx
                    for idx,name in enumerate(list(set(all_names)))
                }
        else:
            self.annotations = [None for s in self.slides]
            self.slide_annotation_names = [None for s in self.slides]

    def gen_slide_patches(self):
         
        # Records-style list of bbox, feature or just bbox if no annotations are passed
        self.data = []
        self.slide_data = []
        # Iterating through
        for s_idx, (slide, pre_slide_annotations, annotation_names) in enumerate(zip(self.slides,self.annotations,self.slide_annotation_names)):
            single_slide_data = {}
            if type(slide)==str:
                slide_tile_source = large_image.open(slide)
            else:
                slide_tile_source = slide

            single_slide_data['image_source'] = slide_tile_source

            slide_metadata = slide_tile_source.getMetadata()
            single_slide_data['metadata'] = slide_metadata
            
            # Find available regions to extract patches from
            if type(self.patch_region)==str:
                if self.patch_region=='tissue':
                    available_regions = self.find_tissue(slide_tile_source, slide_metadata)
                elif self.patch_region=='all':
                    available_regions = {
                        'type': 'FeatureCollection',
                        'features': [{
                            'type': 'Feature',
                            'geometry': {
                                'type': 'Polygon',
                                'coordinates': [[
                                    [0,0],
                                    [0,slide_metadata['sizeY']],
                                    [slide_metadata['sizeX'],slide_metadata['sizeY']],
                                    [slide_metadata['sizeX'],0],
                                    [0,0]
                                ]]
                            },
                            'properties': {}
                        }]
                    }

            elif type(self.patch_region)==list:
                if type(self.patch_region[s_idx])==str:
                    if self.patch_region[s_idx]=='tissue':
                        available_regions = self.find_tissue(slide_tile_source,slide_metadata)
                    elif self.patch_region[s_idx]=='all':
                        available_regions = {
                            'type': 'FeatureCollection',
                            'features': [{
                                'type': 'Feature',
                                'geometry': {
                                    'type': 'Polygon',
                                    'coordinates': [[
                                        [0,0],
                                        [0,slide_metadata['sizeY']],
                                        [slide_metadata['sizeX'],slide_metadata['sizeY']],
                                        [slide_metadata['sizeX'],0],
                                        [0,0]
                                    ]]
                                },
                                'properties': {}
                            }]
                        }
                elif type(self.patch_region[s_idx])==dict:
                    available_regions = self.patch_region[s_idx]

            elif type(self.patch_region)==dict:
                available_regions = self.patch_region

            single_slide_data['available_regions'] = available_regions

            # Finding current annotations
            if not pre_slide_annotations is None:

                # Filtering annotations by available_regions
                if type(pre_slide_annotations)==dict:
                    available_slide_annotations = gpd.sjoin(
                        left_df = pre_slide_annotations,
                        right_df = gpd.GeoDataFrame.from_features(available_regions['features']),
                        how = 'inner',
                        predicate = 'intersects'
                    )

                    available_geo_dict = available_slide_annotations.to_geo_dict(show_bbox=True)
                    available_geo_dict['properties'] = {'name': annotation_names}
                    available_slide_annotations = [available_geo_dict]

                elif type(pre_slide_annotations)==list:
                    available_regions_gdf = gpd.GeoDataFrame.from_features(available_regions['features'])
                    available_slide_annotations = []
                    for s,a in zip(pre_slide_annotations,annotation_names):
                        filtered_s = gpd.sjoin(
                            left_df = s,
                            right_df = available_regions_gdf,
                            how = 'inner',
                            predicate = 'intersects'
                        )

                        filtered_geo_dict = filtered_s.to_geo_dict(show_bbox=True)
                        filtered_geo_dict['properties'] = {'name':a}
                        available_slide_annotations.append(
                            filtered_geo_dict
                        )

                # If spatial or property filters applied
                filtered_slide_annotations, filter_reference_list = process_filters_queries(self.property_filters, self.spatial_filters, [self.use_structures], available_slide_annotations)

                single_slide_data['filtered_annotations'] = filter_reference_list
            else:
                filtered_slide_annotations = {'features': []}
                single_slide_data['filtered_annotations'] = []

            self.slide_data.append(single_slide_data)
            slide_annotations_gdf = gpd.GeoDataFrame.from_features(filtered_slide_annotations['features'])
            # Finding patch coordinates
            if self.patch_mode=='all':
                available_bbox = gpd.GeoDataFrame.from_features(available_regions['features']).total_bounds

                x_start = np.maximum(int(self.patch_size[0]/2),int(available_bbox[0]+(self.patch_size[0]/2)))
                y_start = np.maximum(int(self.patch_size[1]/2),int(available_bbox[1]+(self.patch_size[1]/2)))
                n_x = floor((available_bbox[2]-available_bbox[0])/self.patch_size[0])
                n_y = floor((available_bbox[3]-available_bbox[1])/self.patch_size[1])

                bbox_list = []
                for x in range(0,n_x):
                    for y in range(0,n_y):

                        # Adding column of bboxes
                        bbox = [
                            int(x_start-(self.patch_size[0]/2)),
                            int(y_start-(self.patch_size[1]/2)),
                            int(x_start+(self.patch_size[0]/2)),
                            int(y_start+(self.patch_size[1]/2))
                        ]
                        bbox_list.append(bbox)
                        y_start+=self.patch_size[1]

                    bottom_row_bbox = [
                        int(x_start - (self.patch_size[0]/2)),
                        int(available_bbox[3]-self.patch_size[1]),
                        int(x_start + (self.patch_size[0]/2)),
                        int(available_bbox[3])
                    ]
                    bbox_list.append(bottom_row_bbox)
                    x_start += self.patch_size[0]
                    y_start = np.maximum(self.patch_size[1]/2,available_bbox[1]-(self.patch_size[1]/2))

                # Adding right-side column
                for y in range(0,n_y):
                    right_column_bbox = [
                        int(available_bbox[2] - self.patch_size[0]),
                        int(y_start - (self.patch_size[1]/2)),
                        int(available_bbox[2]),
                        int(y_start + (self.patch_size[1]/2))
                    ]
                    bbox_list.append(right_column_bbox)
                    y_start+=self.patch_size[1]

                # Adding bottom-right corner
                bottom_right_bbox = [
                    int(available_bbox[2]-self.patch_size[0]),
                    int(available_bbox[3]-self.patch_size[1]),
                    int(available_bbox[2]),
                    int(available_bbox[3])
                ]
                bbox_list.append(bottom_right_bbox)

            if self.patch_mode=='overlap':
                #TODO: add overlap patch_mode
                raise NotImplementedError

            if self.patch_mode== 'even':
                # 'even' means all patches within available_regions (with and without intersection with features)
                # Starting out with the bounding box of available regions
                available_bbox = gpd.GeoDataFrame.from_features(available_regions['features']).total_bounds

                # Assuming non-overlapping patches equally dispersed within this box (excludes bottom and right edge)
                x_coords = np.linspace(
                    start = int(available_bbox[0]+(self.patch_size[0]/2)),
                    stop = int(available_bbox[2]-(self.patch_size[0]/2)),
                    num = floor((available_bbox[2]-available_bbox[0])/self.patch_size[0])
                ).tolist()
                y_coords = np.linspace(
                    start = int(available_bbox[1]+(self.patch_size[1]/2)),
                    stop = int(available_bbox[3]-(self.patch_size[1]/2)),
                    num = floor((available_bbox[3]-available_bbox[1])/self.patch_size[1])
                ).tolist()
                # Adding last coordinate
                x_coords.append(np.maximum(int(available_bbox[0]-(self.patch_size[0]/2)),0+(self.patch_size[0]/2)))
                y_coords.append(np.maximum(int(available_bbox[1]-(self.patch_size[1]/2)),0+(self.patch_size[1]/2)))

                # Iterating through both and adding to self.data
                bbox_list = []
                for x in x_coords:
                    for y in y_coords:
                        bbox = [
                            int(x-(self.patch_size[0]/2)),
                            int(y-(self.patch_size[1]/2)),
                            int(x+(self.patch_size[0]/2)),
                            int(y+(self.patch_size[1]/2))
                        ]
                        bbox = [np.maximum(i,0) for i in bbox]
                        bbox_list.append(bbox)

            elif self.patch_mode=='bbox':
                # 'bbox' means each patch will be formed from the bbox of each feature (structure) (patches will initially be different sizes)
                bbox_list = []
                for i in filtered_slide_annotations['features']:
                    bbox = i['bbox']
                    bbox_list.append(bbox)

            elif self.patch_mode=='centered_bbox':
                # 'centered_bbox' means each patch uses the center of the bbox of each feature and expands out to "patch_size"
                centroids = [[(i['bbox'][0]+i['bbox'][2])/2, (i['bbox'][1]+i['bbox'][3])/2] for i in filtered_slide_annotations['features']]
                bbox_list = []
                for c in centroids:
                    bbox = [int(c[0]-(self.patch_size[0]/2)),int(c[1]-(self.patch_size[1]/2)),int(c[0]+(self.patch_size[0]/2)), int(c[1]+(self.patch_size[1]/2))]
                    bbox = [np.maximum(i,0) for i in bbox]
                    bbox_list.append(bbox)

            elif self.patch_mode=='random_bbox':
                # 'random_bbox' means each patch uses the center of the bbox of each feature and expands out to a random amount +/- 0.25 patch_size (patches will initially be different sizes)
                centroids = [[(i['bbox'][0]+i['bbox'][2])/2, (i['bbox'][1]+i['bbox'][3])/2] for i in filtered_slide_annotations['features']]
                width_list = [np.random.randint(int(self.patch_size[0]-(0.25*self.patch_size[0])),int(self.patch_size[0]+(0.25*self.patch_size[0]))) for i in range(len(centroids))]
                height_list = [np.random.randint(int(self.patch_size[1]-(0.25*self.patch_size[1])),int(self.patch_size[1]+(0.25*self.patch_size[1]))) for i in range(len(centroids))]

                bbox_list = []
                for c,w,h in zip(centroids,width_list,height_list):
                    bbox = [int(c[0]-(w/2)),int(c[1]-(h/2)),int(c[0]+(w/2)),int(c[1]+(h/2))]
                    bbox = [np.maximum(i,0) for i in bbox]
                    
                    bbox_list.append(bbox)

            if self.use_parallel:
                self.data.extend(Parallel(
                    n_jobs = self.n_jobs,
                    verbose = 50 if self.verbose else 0,
                    backend = 'threading',
                    return_as = 'list'
                )(
                    delayed(
                        self.make_patch
                    )(i,s_idx,slide_annotations_gdf)
                    for i in bbox_list
                ))
            else:
                if not self.verbose:
                    for i in bbox_list:
                        self.data.append(
                            self.make_patch(i,s_idx,slide_annotations_gdf)
                        )
                else:
                    for i in tqdm(bbox_list):
                        self.data.append(
                            self.make_patch(i,s_idx,slide_annotations_gdf)
                        )

        if self.shuffle:
            random.shuffle(self.data)

    def make_patch(self, bbox:list, slide_idx:int, annotations:gpd.GeoDataFrame):
        
        if not annotations.empty:
            features = annotations[annotations.intersects(box(*bbox))].to_geo_dict(show_bbox=True)['features']
        else:
            features = []

        if not self.use_cache:
            return_dict = {
                'bbox': bbox,
                'features': features,
                'slide_idx': slide_idx
            }
        else:
            image, mask = self.make_image_and_mask(bbox,slide_idx,features)

            return_dict = {
                'bbox': bbox,
                'features': features,
                'slide_idx': slide_idx,
                'image': image,
                'mask': mask
            }

        return return_dict

    def find_tissue(self, slide_tile_source, slide_metadata):
        
        # This process has variable success and most likely will not work the same for all tissues
        if not 'frames' in slide_metadata:
                # Grabbing the thumbnail of the image (RGB)
                thumbnail_img,_ = slide_tile_source.getThumbnail()
                thumb_array = np.array(Image.open(BytesIO(thumbnail_img)))

        else:
            # Getting the max projection of the thumbnail
            thumb_frame_list = []
            for f in range(len(slide_metadata['frames'])):
                thumb,_ = slide_tile_source.getThumbnail()
                thumb = np.array(Image.open(BytesIO(thumb)))
                thumb_frame_list.append(np.max(thumb,axis=-1)[:,:,None])

            thumb_array = np.concatenate(tuple(thumb_frame_list),axis=-1)

        print(f'shape of thumbnail array: {np.shape(thumb_array)}')
        # Getting scale factors for thumbnail image to full-size image
        thumbX, thumbY = np.shape(thumb_array)[1],np.shape(thumb_array)[0]
        scale_x = slide_metadata['sizeX']/thumbX
        scale_y = slide_metadata['sizeY']/thumbY

        thumb_array = 255-thumb_array

        # Mean of all channels/frames to make grayscale mask
        gray_mask = np.squeeze(np.mean(thumb_array,axis=-1))
        print(f'shape of grayscale mask: {np.shape(gray_mask)}')

        threshold_val = threshold_otsu(gray_mask)
        tissue_mask = gray_mask <= threshold_val

        print(f'threshold: {threshold_val}')
        tissue_mask = remove_small_holes(tissue_mask,area_threshold=150)

        labeled_mask = label(tissue_mask)
        tissue_pieces = np.unique(labeled_mask).tolist()
        print(f'Found: {len(tissue_pieces)-1} tissue pieces!')
        tissue_shape_list = []
        for piece in tissue_pieces[1:]:
            tissue_contours = find_contours(labeled_mask==piece)

            for contour in tissue_contours:

                poly_list = [(i[1]*scale_x,i[0]*scale_y) for i in contour]
                if len(poly_list)>2:
                    obj_polygon = Polygon(poly_list)

                    if not obj_polygon.is_valid:
                        made_valid = make_valid(obj_polygon)

                        if made_valid.geom_type=='Polygon':
                            tissue_shape_list.append(made_valid)
                        elif made_valid.geom_type in ['MultiPolygon','GeometryCollection']:
                            for g in made_valid.geoms:
                                if g.geom_type=='Polygon':
                                    tissue_shape_list.append(g)
                    else:
                        tissue_shape_list.append(obj_polygon)

        # Merging shapes together to remove holes
        merged_tissue = unary_union(tissue_shape_list)
        if merged_tissue.geom_type=='Polygon':
            merged_tissue = [merged_tissue]
        elif merged_tissue.geom_type in ['MultiPolygon','GeometryCollection']:
            merged_tissue = merged_tissue.geoms

        thumbnail_geojson = {
            'type': 'FeatureCollection',
            'features': [
                {'type':'Feature','geometry': {'type': 'Polygon','coordinates': [list(i.exterior.coords)]}}
                for i in merged_tissue if i.geom_type=='Polygon'
            ]
        }

        return thumbnail_geojson

    def make_image_and_mask(self, bbox:list, slide_idx:int, features: list):
        
        image_source = self.slide_data[slide_idx]['image_source']
        if not 'frames' in self.slide_data[slide_idx]['metadata']:
            image,_ = image_source.getRegion(
                format = large_image.constants.TILE_FORMAT_NUMPY,
                region = {
                    'left': bbox[0],
                    'top': bbox[1],
                    'right': bbox[2],
                    'bottom': bbox[3]
                }
            )
        else:
            image = np.zeros((int(bbox[3]-bbox[1]),int(bbox[2]-bbox[0]),len(self.slide_data[slide_idx]['metadata']['frames'])))
            
            for f in range(len(self.slide_data[slide_idx]['metadata']['frames'])):
                image_frame,_ = image_source.getRegion(
                    format = large_image.constants.TILE_FORMAT_NUMPY,
                    region = {
                        'left': bbox[0],
                        'top': bbox[1],
                        'right': bbox[2],
                        'bottom': bbox[3]
                    },
                    frame = f
                )
                image[:,:,f] += np.squeeze(image_frame)


        if len(features)>0:
            mask = np.zeros((int(bbox[3]-bbox[1]),int(bbox[2]-bbox[0]),len(list(self.mask_key.keys()))))
            bbox_box = box(*bbox)
            height = int(bbox[3]-bbox[1])
            width = int(bbox[2]-bbox[0])
            for f in features:
                feature_shape = shape(f['geometry'])
                bbox_intersection = feature_shape.intersection(bbox_box)

                if bbox_intersection.area>0 and bbox_intersection.geom_type=='Polygon':
                    int_coords = np.array(list(bbox_intersection.exterior.coords))
                    min_x = bbox[0]
                    min_y = bbox[1]

                    scaled_coords = np.flip(np.squeeze(int_coords) - np.array([min_x, min_y]),axis=1)
                    f_mask = polygon2mask(
                        image_shape = (height,width),
                        polygon = scaled_coords
                    ).astype(int)

                    mask[:,:,self.mask_key[f['properties']['name']]] += f_mask
                
                elif bbox_intersection.area>0 and bbox_intersection.geom_type=='GeometryCollection':
                    for g in bbox_intersection.geoms:
                        if g.geom_type=='Polygon':
                            int_coords = np.array(list(g.exterior.coords))
                            min_x = bbox[0]
                            min_y = bbox[1]

                            scaled_coords = np.flip(np.squeeze(int_coords) - np.array([min_x, min_y]),axis=1)
                            f_mask = polygon2mask(
                                image_shape = (height,width),
                                polygon = scaled_coords
                            ).astype(int)

                            mask[:,:,self.mask_key[f['properties']['name']]] += f_mask

        else:
            mask = np.zeros((int(bbox[3]-bbox[1]),int(bbox[2]-bbox[0])))


        return image, mask

    def get_next_image(self, idx:int):
        
        next_data = self.data[idx]

        if self.use_cache:
            image = next_data['image']
            mask = next_data['mask']

        else:
            image, mask = self.make_image_and_mask(next_data['bbox'], next_data['slide_idx'], next_data['features'])
        
        return image, mask

    def __len__(self):
        return len(self.data)

    def __getitem__(self,idx):
        
        image, mask = self.get_next_image(idx)

        if self.transforms:
            image = self.transforms(image)

        if self.target_transforms:
            mask = self.target_transforms(mask)

        return image, mask

    def __str__(self):
        
        key_configs = {
            'slide_data': [
                {i:j for i,j in k.items() if not i=='image_source'}
                for k in self.slide_data
            ],
            'patches': {
                'region': self.patch_region,
                'mode': self.patch_mode,
                'size': self.patch_size
            },
            'property_filters': self.property_filters,
            'spatial_filters': self.spatial_filters,
            'mask_key': self.mask_key,
            'use_structures': self.use_structures,
            'use_cache': self.use_cache,
            'shuffle': self.shuffle,
            'seed_val': self.seed_val
        }

        return json.dumps(key_configs)

    def export_configs(self, save_path: str):
        
        # mask_key, patch_size, patch_mode, patch_region
        # spatial filters, property_filters, use_structures, use_cache, shuffle, seed_val
        key_configs = {
            'slide_data': [
                {i:j for i,j in k.items() if not i=='image_source'}
                for k in self.slide_data
            ],
            'patches': {
                'region': self.patch_region,
                'mode': self.patch_mode,
                'size': self.patch_size
            },
            'property_filters': self.property_filters,
            'spatial_filters': self.spatial_filters,
            'mask_key': self.mask_key,
            'use_structures': self.use_structures,
            'use_cache': self.use_cache,
            'shuffle': self.shuffle,
            'seed_val': self.seed_val
        }

        with open(save_path,'w') as f:
            json.dump(key_configs, f, indent=4)


class ClassificationDataset:
    def __init__(self,
                 slides: list = [],
                 annotations: Union[list,dict,None] = None,
                 label_property: Union[list,str,None] = None,
                 transforms: Union[None, Callable] = None,
                 label_transforms: Union[None, Callable] = None,
                 patch_size: list = [256,256],
                 patch_mode: str = 'bbox',
                 patch_region: Union[str,list,dict] = 'all',
                 spatial_filters: Union[list,None] = None,
                 property_filters: Union[list,None] = None,
                 use_structures: Union[str,list] = 'all',
                 use_cache: bool = True,
                 shuffle: bool = True,
                 seed_val: int = 1701,
                 verbose: bool = False,
                 use_parallel: bool = True,
                 n_jobs: int = 1
                 ):
        
        self.slides = slides
        self.pre_annotations = annotations
        self.label_property = label_property
        self.transforms = transforms
        self.label_transforms = label_transforms

        self.patch_size = patch_size
        self.patch_mode = patch_mode
        if self.pre_annotations is None:
            if 'bbox' in self.patch_mode:
                raise AssertionError('If no annotations are passed, patch_mode must be "all"')
        self.patch_region = patch_region
        if type(self.patch_region)==str:
            assert self.patch_region in ['all','tissue']
        elif type(self.patch_region)==list:
            assert all([any([i==j for j in ['all','tissue']]) if type(i)==str else type(i)==dict for i in self.patch_region])
        self.spatial_filters = spatial_filters if not spatial_filters is None else []
        self.property_filters = property_filters if not property_filters is None else []
        
        self.use_structures = use_structures
        self.use_cache = use_cache
        self.shuffle = shuffle
        self.seed_val = seed_val

        # Setting seed for reproducibility (🖖)
        self.seed_val = seed_val
        self.plant_seed(seed_val)
        self.verbose = verbose
        self.use_parallel = use_parallel
        self.n_jobs = n_jobs
        self.process_annotations()
        self.gen_slide_patches()

    def plant_seed(self, seed_val):

        random.seed(seed_val)
        os.environ["PYTHONHASHSEED"] = str(seed_val)
        np.random.seed(seed_val)
        #torch.manual_seed(seed_val)
        #torch.cuda.manual_seed(seed_val)
        
    def process_annotations(self):

        # Starting out with assuming all GeoJSON type
        self.slide_annotation_names = []
        if not self.pre_annotations is None:
            if type(self.pre_annotations)==dict:
                self.annotations = [gpd.GeoDataFrame.from_features(self.pre_annotations['features'])]
                self.slide_annotation_names.append(self.pre_annotations['properties']['name'])
            else:
                self.annotations = []
                for a in self.pre_annotations:
                    if type(a)==dict:
                        self.annotations.append([gpd.GeoDataFrame.from_features(a['features'])])
                        self.slide_annotation_names.append([a['properties']['name']])
                    elif type(a)==list:
                        self.annotations.append([gpd.GeoDataFrame.from_features(b['features']) for b in a if type(b)==dict])
                        self.slide_annotation_names.append([b['properties']['name'] for b in a if type(b)==dict])
        else:       
            self.annotations = None

    def gen_slide_patches(self):
        
        # Records-style list of bbox, feature or just bbox if no annotations are passed
        self.data = []
        self.slide_data = []

        # Iterating through
        for s_idx, (slide,pre_slide_annotations,annotation_names) in enumerate(zip(self.slides, self.annotations, self.slide_annotation_names)):
            single_slide_data = {}
            if type(slide)==str:
                slide_tile_source = large_image.open(slide)
            else:
                slide_tile_source = slide

            single_slide_data['image_source'] = slide_tile_source

            slide_metadata = slide_tile_source.getMetadata()
            single_slide_data['metadata'] = slide_metadata
            
            # Find available regions to extract patches from
            if type(self.patch_region)==str:
                if self.patch_region=='tissue':
                    available_regions = self.find_tissue(slide_tile_source,slide_metadata)
                elif self.patch_region=='all':
                    available_regions = {
                        'type': 'FeatureCollection',
                        'features': [{
                            'type': 'Feature',
                            'geometry': {
                                'type': 'Polygon',
                                'coordinates': [[
                                    [0,0],
                                    [0,slide_metadata['sizeY']],
                                    [slide_metadata['sizeX'],slide_metadata['sizeY']],
                                    [slide_metadata['sizeX'],0],
                                    [0,0]
                                ]]
                            },
                            'properties': {}
                        }]
                    }

            elif type(self.patch_region)==list:
                if type(self.patch_region[s_idx])==str:
                    if self.patch_region[s_idx]=='tissue':
                        available_regions = self.find_tissue(slide_tile_source,slide_metadata)
                    elif self.patch_region[s_idx]=='all':
                        available_regions = {
                            'type': 'FeatureCollection',
                            'features': [{
                                'type': 'Feature',
                                'geometry': {
                                    'type': 'Polygon',
                                    'coordinates': [[
                                        [0,0],
                                        [0,slide_metadata['sizeY']],
                                        [slide_metadata['sizeX'],slide_metadata['sizeY']],
                                        [slide_metadata['sizeX'],0],
                                        [0,0]
                                    ]]
                                },
                                'properties': {}
                            }]
                        }
                elif type(self.patch_region[s_idx])==dict:
                    available_regions = self.patch_region[s_idx]

            elif type(self.patch_region)==dict:
                available_regions = self.patch_region

            single_slide_data['available_regions'] = available_regions

            # Finding current annotations
            if not self.annotations is None:

                # Filtering annotations by available_regions
                if type(pre_slide_annotations)==gpd.GeoDataFrame:
                    available_slide_annotations = gpd.sjoin(
                        left_df = pre_slide_annotations,
                        right_df = gpd.GeoDataFrame.from_features(available_regions['features']),
                        how = 'inner',
                        predicate = 'intersects'
                    )

                    available_geo_dict = available_slide_annotations.to_geo_dict(show_bbox=True)
                    available_geo_dict['properties'] = {'name': annotation_names}
                    available_slide_annotations = [available_geo_dict]

                elif type(pre_slide_annotations)==list:
                    available_regions_gdf = gpd.GeoDataFrame.from_features(available_regions['features'])
                    available_slide_annotations = []
                    for s,a in zip(pre_slide_annotations,annotation_names):
                        filtered_s = gpd.sjoin(
                            left_df = s,
                            right_df = available_regions_gdf,
                            how = 'inner',
                            predicate = 'intersects'
                        )

                        filtered_geo_dict = filtered_s.to_geo_dict(show_bbox=True)
                        filtered_geo_dict['properties'] = {'name': a}
                        available_slide_annotations.append(
                            filtered_geo_dict
                        )


                # If spatial or property filters applied
                filtered_slide_annotations, filter_reference_list = process_filters_queries(self.property_filters, self.spatial_filters, [self.use_structures], available_slide_annotations)

                single_slide_data['filtered_annotations'] = filter_reference_list
            else:
                filtered_slide_annotations = {'features': []}
                single_slide_data['filtered_annotations'] = []

            self.slide_data.append(single_slide_data)      
            slide_annotations_gdf = gpd.GeoDataFrame.from_features(filtered_slide_annotations['features'])
            # Finding patch coordinates
            if self.patch_mode== 'all':
                # 'all' means all patches within available_regions (with and without intersection with features)
                # Starting out with the bounding box of available regions
                available_bbox = gpd.GeoDataFrame.from_features(available_regions['features']).total_bounds

                # Assuming non-overlapping patches equally dispersed within this box (excludes bottom and right edge)
                x_coords = np.linspace(
                    start = int(available_bbox[0]+(self.patch_size[0]/2)),
                    stop = int(available_bbox[2]-(self.patch_size[0]/2)),
                    num = floor((available_bbox[2]-available_bbox[0])/self.patch_size[0])
                ).tolist()
                y_coords = np.linspace(
                    start = int(available_bbox[1]+(self.patch_size[1]/2)),
                    stop = int(available_bbox[3]-(self.patch_size[1]/2)),
                    num = floor((available_bbox[3]-available_bbox[1])/self.patch_size[1])
                ).tolist()
                # Adding last coordinate
                x_coords.append(np.maximum(int(available_bbox[0]-(self.patch_size[0]/2)),0+(self.patch_size[0]/2)))
                y_coords.append(np.maximum(int(available_bbox[1]-(self.patch_size[1]/2)),0+(self.patch_size[1]/2)))

                # Iterating through both and adding to self.data
                bbox_list = []
                for x in x_coords:
                    for y in y_coords:
                        bbox = [
                            int(x-(self.patch_size[0]/2)),
                            int(y-(self.patch_size[1]/2)),
                            int(x+(self.patch_size[0]/2)),
                            int(y+(self.patch_size[1]/2))
                        ]
                        bbox = [np.maximum(i,0) for i in bbox]
                        bbox_list.append(bbox)

            elif self.patch_mode=='bbox':
                # 'bbox' means each patch will be formed from the bbox of each feature (structure) (patches will initially be different sizes)
                bbox_list = []
                for i in filtered_slide_annotations['features']:
                    bbox = i['bbox']
                    bbox_list.append(bbox)

            elif self.patch_mode=='centered_bbox':
                # 'centered_bbox' means each patch uses the center of the bbox of each feature and expands out to "patch_size"
                centroids = [[(i['bbox'][0]+i['bbox'][2])/2, (i['bbox'][1]+i['bbox'][3])/2] for i in filtered_slide_annotations['features']]

                bbox_list = []
                for c in centroids:
                    bbox = [int(c[0]-(self.patch_size[0]/2)),int(c[1]-(self.patch_size[1]/2)),int(c[0]+(self.patch_size[0]/2)), int(c[1]+(self.patch_size[1]/2))]
                    bbox = [np.maximum(i,0) for i in bbox]

                    bbox_list.append(bbox)

            elif self.patch_mode=='random_bbox':
                # 'random_bbox' means each patch uses the center of the bbox of each feature and expands out to a random amount +/- 0.25 patch_size (patches will initially be different sizes)
                centroids = [[(i['bbox'][0]+i['bbox'][2])/2, (i['bbox'][1]+i['bbox'][3])/2] for i in filtered_slide_annotations['features']]
                width_list = [np.random.randint(int(self.patch_size[0]-(0.25*self.patch_size[0])),int(self.patch_size[0]+(0.25*self.patch_size[0]))) for i in range(len(centroids))]
                height_list = [np.random.randint(int(self.patch_size[1]-(0.25*self.patch_size[1])),int(self.patch_size[1]+(0.25*self.patch_size[1]))) for i in range(len(centroids))]

                bbox_list = []
                for c,w,h in zip(centroids,width_list,height_list):
                    bbox = [int(c[0]-(w/2)),int(c[1]-(h/2)),int(c[0]+(w/2)),int(c[1]+(h/2))]
                    bbox = [np.maximum(i,0) for i in bbox]

                    bbox_list.append(bbox)

            if self.use_parallel:
                self.data.extend(Parallel(
                    n_jobs = self.n_jobs,
                    verbose = 50 if self.verbose else 0,
                    backend = 'threading',
                    return_as = 'list'
                )(
                    delayed(
                        self.make_patch
                    )(i,s_idx,slide_annotations_gdf)
                    for i in bbox_list
                ))
            else:
                if not self.verbose:
                    for i in bbox_list:
                        self.data.append(
                            self.make_patch(i,s_idx,slide_annotations_gdf)
                        )
                else:
                    for i in tqdm(bbox_list):
                        self.data.append(
                            self.make_patch(i,s_idx,slide_annotations_gdf)
                        )

        if self.shuffle:
            random.shuffle(self.data)

    def make_patch(self,bbox:list, slide_idx: int, annotations:gpd.GeoDataFrame):

        features = annotations[annotations.intersects(box(*bbox))].to_geo_dict(show_bbox=True)['features']

        if not self.use_cache:
            return_dict = {
                'bbox': bbox,
                'features': features,
                'slide_idx': slide_idx
            }
        else:
            image, label = self.make_image_and_label(bbox,slide_idx,features)

            return_dict = {
                'bbox': bbox,
                'features': features,
                'slide_idx': slide_idx,
                'image': image,
                'label': label
            }

        return return_dict

    def make_image_and_label(self, bbox:list, slide_idx:int, features: list):

        image_source = self.slide_data[slide_idx]['image_source']
        if not 'frames' in self.slide_data[slide_idx]['metadata']:
            image,_ = image_source.getRegion(
                format = large_image.constants.TILE_FORMAT_NUMPY,
                region = {
                    'left': bbox[0],
                    'top': bbox[1],
                    'right': bbox[2],
                    'bottom': bbox[3]
                }
            )
        else:
            image = np.zeros((int(bbox[3]-bbox[1]),int(bbox[2]-bbox[0]),len(self.slide_data[slide_idx]['metadata']['frames'])))
            for f in range(len(self.slide_data[slide_idx]['metadata']['frames'])):
                image_frame,_ = image_source.getRegion(
                    format = large_image.constants.TILE_FORMAT_NUMPY,
                    region = {
                        'left': bbox[0],
                        'top': bbox[1],
                        'right': bbox[2],
                        'bottom': bbox[3]
                    },
                    frame = f
                )
                image[:,:,f] += image_frame

        # For now this will return multiple "labels" depending on the number of structures that intersect with this bbox, label_transforms can prune this if needed
        label = []
        for f in features:
            if type(self.label_property)==str:
                label_list = [self.label_property]
            elif type(self.label_property)==list:
                label_list = self.label_property

            for l_p in label_list:
                if not '-->' in l_p:
                    if l_p in f['properties']:
                        try:
                            label.append(float(f['properties'][l_p]))
                        except ValueError:
                            label.append(f['properties'][l_p])
                else:
                    l_parts = l_p.split(' --> ')
                    f_props_copy = f['properties'].copy()
                    for l in l_parts:
                        if not f_props_copy is None:
                            if l in f_props_copy:
                                f_props_copy = f_props_copy[l]
                            else:
                                f_props_copy = None
                    
                    if not f_props_copy is None:
                        try:
                            label.append(float(f_props_copy))
                        except ValueError:
                            label.append(f_props_copy)
                    else:
                        label.append(float(0))
        
        return image, label

    def get_next_image(self, idx:int):
        
        next_data = self.data[idx]

        if self.use_cache:
            image = next_data['image']
            label = next_data['label']

        else:
            image, label = self.make_image_and_label(next_data['bbox'], next_data['slide_idx'], next_data['features'])
        
        return image, label

    def __len__(self):
        return len(self.data)

    def __getitem__(self,idx):
        image, label = self.get_next_image(idx)

        if self.transforms:
            image = self.transforms(image)

        if self.label_transforms:
            label = self.label_transforms(label)

        return image, label

    def __str__(self):
        
        key_configs = {
            'slide_data': [
                {i:j for i,j in k.items() if not i=='image_source'}
                for k in self.slide_data
            ],
            'patches': {
                'region': self.patch_region,
                'mode': self.patch_mode,
                'size': self.patch_size
            },
            'property_filters': self.property_filters,
            'spatial_filters': self.spatial_filters,
            'label_property': self.label_property,
            'use_structures': self.use_structures,
            'use_cache': self.use_cache,
            'shuffle': self.shuffle,
            'seed_val': self.seed_val
        }

        return json.dumps(key_configs)

    def export_configs(self, save_path:str):
        
        # patch_size, patch_mode, patch_region, label_property
        # spatial filters, property_filters, use_structures, use_cache, shuffle, seed_val
        key_configs = {
            'slide_data': [
                {i:j for i,j in k.items() if not i=='image_source'}
                for k in self.slide_data
            ],
            'patches': {
                'region': self.patch_region,
                'mode': self.patch_mode,
                'size': self.patch_size
            },
            'property_filters': self.property_filters,
            'spatial_filters': self.spatial_filters,
            'label_property': self.label_property,
            'use_structures': self.use_structures,
            'use_cache': self.use_cache,
            'shuffle': self.shuffle,
            'seed_val': self.seed_val
        }

        with open(save_path,'w') as f:
            json.dump(key_configs, f, indent=4)





"""Generating FUSION (Functional Unit State Identification in WSIs) layout
"""
from fusion_tools.visualization import Visualization
from fusion_tools.handler.dsa_handler import DSAHandler
from fusion_tools.components import (
    SlideMap,
    MultiFrameSlideMap,
    ChannelMixer,
    OverlayOptions,
    PropertyViewer,
    GlobalPropertyPlotter,
    HRAViewer,
    DataExtractor,
    BulkLabels
)

from fusion_tools.fusion.data_types import get_upload_types


def get_layout(args):

    # Initialize DSA Handler
    dsa_handler = DSAHandler(
        girderApiUrl=args['girderApiUrl'],
        username = args['user'],
        password=args['pword']
    )

    dsa_login_component = dsa_handler.create_login_component()
    
    dsa_plugin_progress = dsa_handler.create_plugin_progress()

    dsa_dataset_builder = dsa_handler.create_dataset_builder()

    dsa_dataset_uploader = dsa_handler.create_uploader(
        upload_types = get_upload_types()
    )

    user_surveys = []
    #dsa_user_surveys = [dsa_handler.create_survey(i) for i in user_surveys]

    initial_items = args['initialItems']

    fusion_vis = Visualization(
        tileservers=[dsa_handler.get_tile_server(i) for i in initial_items],
        linkage = 'page',
        header = [
            dsa_login_component,
            dsa_plugin_progress,
            #dsa_user_surveys
        ],
        components = {
            "Visualization": [
                [
                    SlideMap(),
                    [
                        OverlayOptions(),
                        PropertyViewer(ignore_list=['_id','_index']),
                        GlobalPropertyPlotter(ignore_list = ['_id','_index']),
                        HRAViewer(),
                        BulkLabels(),
                        DataExtractor()
                    ]
                ]
            ],
            "MultiFrame Visualization": [
                [
                    MultiFrameSlideMap(),
                    [
                        ChannelMixer(),
                        OverlayOptions(),
                        PropertyViewer(ignore_list = ['_id','_index']),
                        GlobalPropertyPlotter(ignore_list = ['_id','_index']),
                        HRAViewer(),
                        BulkLabels(),
                        DataExtractor()
                    ]
                ]
            ],
            "Dataset Builder": [
                dsa_dataset_builder
            ],
            "Dataset Uploader": [
                dsa_dataset_uploader
            ]
        },
        app_options = args['app_options']
    )

    return fusion_vis



""""
Testing property plotter using remote slide with rich annotation properties
"""
import sys
sys.path.append('./src/')
from fusion_tools.visualization import Visualization
from fusion_tools.handler.dsa_handler import DSAHandler
from fusion_tools.components import LargeSlideMap,SlideMap,OverlayOptions, PropertyPlotter, HRAViewer

def main():

    # Grabbing first item from demo DSA instance
    base_url = 'http://ec2-3-230-122-132.compute-1.amazonaws.com:8080/api/v1'
    item_id = '67ae1826fcdeba1e292f7057'

    # Starting the DSAHandler to grab information:
    dsa_handler = DSAHandler(
        girderApiUrl = base_url
    )
    
    vis_session = Visualization(
        tileservers = [dsa_handler.get_tile_server(item_id)],
        components = [
            [
                SlideMap(),
                [
                    OverlayOptions(),
                    PropertyPlotter(),
                    HRAViewer()
                ]
            ]
        ]
    )

    vis_session.start()


if __name__=='__main__':
    main()




"""DSALoginComponent
"""

import json

from typing_extensions import Union

# Dash imports
import dash
dash._dash_renderer._set_react_version('18.2.0')
from dash import callback, ctx, ALL, MATCH, exceptions, no_update
import dash_bootstrap_components as dbc
from dash_extensions.enrich import DashBlueprint, html, Input, Output, State, PrefixIdTransform, MultiplexerTransform

from fusion_tools.visualization.vis_utils import get_pattern_matching_value
from fusion_tools import DSATool


class DSALoginComponent(DSATool):
    """Handler for DSALoginComponent, enabling login to the running DSA instance

    :param DSATool: Sub-class of Tool specific to DSA components. Updates with session data by default.
    :type DSATool: None

    """
    def __init__(self,
                 handler,
                 default_user: Union[dict,None] = None
                ):
        
        super().__init__()
        self.handler = handler
        self.default_user = default_user

        self.modal_size = 'lg'

    def load(self,component_prefix:int):

        self.component_prefix = component_prefix

        self.title = 'DSA Login Component'
        self.blueprint = DashBlueprint(
            transforms=[
                PrefixIdTransform(prefix=f'{component_prefix}'),
                MultiplexerTransform()
            ]
        )

        self.get_callbacks()
    
    def update_layout(self, session_data:dict, use_prefix:bool):
        
                
        layout = html.Div([
            html.H4(
                id = {'type': 'dsa-login-current-user','index': 0},
                children = [
                    f'Welcome, {session_data["current_user"]["login"]}!' if "current_user" in session_data else 'Welcome, Guest!'
                ]
            ),
            html.Hr(),
            html.Div(
                id = {'type': 'dsa-login-div','index': 0},
                children = [
                    dbc.Stack([
                        dbc.Button(
                            'Login',
                            className = 'd-grid col-6 mx-auto',
                            color = 'success',
                            id = {'type': 'dsa-login-button','index': 0}
                        ),
                        dbc.Tooltip(
                            'For registered users, login to view your previous uploads or shared collections!',
                            target = {'type': 'dsa-login-button','index': 0},
                            placement='top'
                        ),
                        dbc.Button(
                            'Create an Account',
                            className = 'd-grid col-6 mx-auto',
                            color = 'warning',
                            id = {'type': 'dsa-login-create-account-button','index': 0}
                        ),
                        dbc.Tooltip(
                            'Create an account in order to upload slides to the DSA instance, to access user surveys, or to share data!',
                            target = {'type': 'dsa-login-create-account-button','index': 0},
                            placement='top'
                        )
                    ],direction='horizontal',gap=2)
                ]
            )
        ],style = {'marginTop':'10px','marginBottom':'10px','marginLeft':'10px','marginRight':'10px'})

        if use_prefix:
            PrefixIdTransform(prefix=self.component_prefix).transform_layout(layout)

        return layout

    def gen_layout(self,session_data:dict):
        """Creating the layout for this component, assigning it to the DashBlueprint object

        :param session_data: Dictionary containing relevant information for the current session
        :type session_data: dict
        """

        self.blueprint.layout = self.update_layout(session_data,use_prefix=False)

    def get_callbacks(self):

        # Callback for selecting Login vs. Create Account
        self.blueprint.callback(
            [
                Input({'type': 'dsa-login-button','index': ALL},'n_clicks'),
                Input({'type': 'dsa-login-create-account-button','index': ALL},'n_clicks'),
                Input({'type': 'dsa-login-back-icon','index': ALL},'n_clicks')
            ],
            [
                Output({'type': 'dsa-login-div','index': ALL},'children')
            ]
        )(self.display_login_fields)

        # Callback for clicking Login button with username and password
        self.blueprint.callback(
            [
                Input({'type': 'dsa-login-login-submit','index': ALL},'n_clicks')
            ],
            [
                State({'type': 'dsa-login-username-input','index': ALL},'value'),
                State({'type': 'dsa-login-password-input','index': ALL},'value'),
                State('anchor-vis-store','data')
            ],
            [
                Output({'type': 'dsa-login-username-error-div','index': ALL},'children'),
                Output({'type': 'dsa-login-password-error-div','index': ALL},'children'),
                Output({'type': 'dsa-login-login-error-div','index':ALL},'children'),
                Output({'type': 'dsa-login-current-user','index': ALL},'children'),
                Output('anchor-vis-store','data')
            ]
        )(self.submit_login)

        # Callback for clicking Create Account with input details
        self.blueprint.callback(
            [
                Input({'type':'dsa-login-create-account-submit','index': ALL},'n_clicks')
            ],
            [
                State({'type': 'dsa-login-firstname-input','index': ALL},'value'),
                State({'type': 'dsa-login-lastname-input','index': ALL},'value'),
                State({'type': 'dsa-login-email-input','index': ALL},'value'),
                State({'type': 'dsa-login-username-input','index': ALL},'value'),
                State({'type': 'dsa-login-password-input','index': ALL},'value'),
                State('anchor-vis-store','data')
            ],
            [
                Output({'type': 'dsa-login-email-error-div','index': ALL},'children'),
                Output({'type': 'dsa-login-username-error-div','index': ALL},'children'),
                Output({'type': 'dsa-login-password-error-div','index': ALL},'children'),
                Output({'type': 'dsa-login-create-account-error-div','index': ALL},'children'),
                Output({'type': 'dsa-login-current-user','index': ALL},'children'),
                Output('anchor-vis-store','data')
            ]
        )(self.submit_create_account)
        
    def display_login_fields(self, login_clicked, create_account_clicked, back_clicked):
        """Displaying login fields depending on if login, create account, or back is clicked

        :param login_clicked: Login button clicked
        :type login_clicked: list
        :param create_account_clicked: Create Account button clicked
        :type create_account_clicked: list
        :param back_clicked: Back icon clicked
        :type back_clicked: list
        """

        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate

        if 'dsa-login-button' in ctx.triggered_id['type']:
            # Create input fields for "username" and "password" and a button for Login
            new_children = html.Div([
                dbc.Row([
                    dbc.Col(
                        html.A(
                            dbc.Stack([
                                html.P(
                                    html.I(
                                        className = 'fa-solid fa-arrow-left',
                                        style = {'marginRight':'5px'}
                                    )
                                ),
                                html.P(
                                    'Back'
                                )
                            ],direction='horizontal'),
                            id = {'type': f'{self.component_prefix}-dsa-login-back-icon','index': 0},
                            n_clicks = 0
                        )
                    )
                ]),
                dbc.Stack([
                    dbc.InputGroup([
                        dbc.InputGroupText(
                            'Username: '
                        ),
                        dbc.Input(
                            id = {'type': f'{self.component_prefix}-dsa-login-username-input','index': 0},
                            placeholder='username',
                            type = 'text',
                            required = True,
                            value = [],
                            maxLength = 1000,
                            pattern = '^[a-z0-9]+$'
                        )
                    ]),
                    html.Div(
                        id = {'type': f'{self.component_prefix}-dsa-login-username-error-div','index': 0},
                        children = []
                    ),
                    dbc.InputGroup([
                        dbc.InputGroupText(
                            'Password: '
                        ),
                        dbc.Input(
                            id = {'type': f'{self.component_prefix}-dsa-login-password-input','index': 0},
                            type = 'password',
                            required = True,
                            value = [],
                            maxLength = 1000
                        )
                    ]),
                    html.Div(
                        id = {'type': f'{self.component_prefix}-dsa-login-password-error-div','index': 0},
                        children = []
                    ),
                    dbc.Button(
                        'Login!',
                        className = 'd-grid col-12 mx-auto',
                        color = 'primary',
                        id = {'type': f'{self.component_prefix}-dsa-login-login-submit','index': 0},
                        n_clicks = 0
                    ),
                    html.Div(
                        id = {'type': f'{self.component_prefix}-dsa-login-login-error-div','index': 0},
                        children = []
                    )
                ],direction = 'vertical',gap=1)
            ])

        elif 'dsa-login-create-account-button' in ctx.triggered_id['type']:
            # Create input fields for "username" and "password" and a button for Login
            new_children = html.Div([
                dbc.Row([
                    dbc.Col(
                        html.A(
                            dbc.Stack([
                                html.P(
                                    html.I(
                                        className = 'fa-solid fa-arrow-left',
                                        style = {'marginRight':'2px'}
                                    )
                                ),
                                html.P(
                                    'Back'
                                )
                            ],direction='horizontal'),
                            id = {'type': f'{self.component_prefix}-dsa-login-back-icon','index': 0},
                            n_clicks = 0
                        )
                    )
                ]),
                dbc.Stack([
                    dbc.InputGroup([
                        dbc.InputGroupText(
                            'First Name: '
                        ),
                        dbc.Input(
                            id = {'type': f'{self.component_prefix}-dsa-login-firstname-input','index': 0},
                            placeholder='First Name',
                            type = 'text',
                            required = True,
                            value = [],
                            maxLength = 1000,
                            pattern = '^[a-zA-Z]+$'
                        )
                    ]),
                    dbc.InputGroup([
                        dbc.InputGroupText(
                            'Last Name: '
                        ),
                        dbc.Input(
                            id = {'type': f'{self.component_prefix}-dsa-login-lastname-input','index':0},
                            placeholder = 'Last Name',
                            type = 'text',
                            required = True,
                            value = [],
                            maxLength = 1000,
                            pattern = '^[a-zA-Z]+$'
                        )
                    ]),
                    dbc.InputGroup([
                        dbc.InputGroupText(
                            'Email: '
                        ),
                        dbc.Input(
                            id = {'type': f'{self.component_prefix}-dsa-login-email-input','index':0},
                            placeholder = 'email address',
                            type = 'email',
                            required = True,
                            value = [],
                            maxLength = 1000,
                            pattern = "^[a-zA-Z0-9.!#$%&'*+/=?^_`{|}~-]+@[a-zA-Z0-9-]+(?:\.[a-zA-Z0-9-]+)*$"
                        )
                    ]),
                    html.Div(
                        id = {'type': f'{self.component_prefix}-dsa-login-email-error-div','index': 0},
                        children = []
                    ),
                    dbc.InputGroup([
                        dbc.InputGroupText(
                            'Username: '
                        ),
                        dbc.Input(
                            id = {'type': f'{self.component_prefix}-dsa-login-username-input','index': 0},
                            placeholder='username',
                            type = 'text',
                            required = True,
                            value = [],
                            maxLength = 1000,
                            pattern = '^[a-z0-9]+$'
                        )
                    ]),
                    html.Div(
                        id = {'type': f'{self.component_prefix}-dsa-login-username-error-div','index': 0},
                        children = []
                    ),
                    dbc.InputGroup([
                        dbc.InputGroupText(
                            'Password: '
                        ),
                        dbc.Input(
                            id = {'type': f'{self.component_prefix}-dsa-login-password-input','index': 0},
                            type = 'password',
                            required = True,
                            value = [],
                            maxLength = 1000
                        )
                    ]),
                    html.Div(
                        id = {'type': f'{self.component_prefix}-dsa-login-password-error-div','index': 0},
                        children = []
                    ),
                    dbc.Button(
                        'Login!',
                        className = 'd-grid col-12 mx-auto',
                        color = 'primary',
                        id = {'type': f'{self.component_prefix}-dsa-login-create-account-submit','index': 0},
                        n_clicks = 0
                    ),
                    html.Div(
                        id = {'type': f'{self.component_prefix}-dsa-login-create-account-error-div','index':0},
                        children = []
                    )
                ],direction='vertical',gap=1)
            ])
    
        elif 'dsa-login-back-icon' in ctx.triggered_id['type']:
            new_children = dbc.Stack([
                dbc.Button(
                    'Login',
                    className = 'd-grid col-6 mx-auto',
                    color = 'success',
                    id = {'type': f'{self.component_prefix}-dsa-login-button','index': 0}
                ),
                dbc.Tooltip(
                    'For registered users, login to view your previous uploads or shared collections!',
                    target = {'type': f'{self.component_prefix}-dsa-login-button','index': 0},
                    placement='top'
                ),
                dbc.Button(
                    'Create an Account',
                    className = 'd-grid col-6 mx-auto',
                    color = 'warning',
                    id = {'type': f'{self.component_prefix}-dsa-login-create-account-button','index': 0}
                ),
                dbc.Tooltip(
                    'Create an account in order to upload slides to the DSA instance, to access user surveys, or to share data!',
                    target = {'type': f'{self.component_prefix}-dsa-create-account-button','index': 0},
                    placement='top'
                )
            ],direction='horizontal',gap=2)

        return [new_children]
    
    def submit_login(self, login_clicked,username_input, password_input, session_data):
        
        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        session_data = json.loads(session_data)
        
        username_input = get_pattern_matching_value(username_input)
        password_input = get_pattern_matching_value(password_input)

        if username_input is None or username_input == '':
            username_error_div = dbc.Alert('Make sure to enter a username!',color = 'danger')
        else:
            username_error_div = []
        
        if password_input is None or password_input == '':
            password_error_div = dbc.Alert('Make sure to enter your password!',color = 'danger')
        else:
            password_error_div = []

        if not any([i is None or i=='' for i in [username_input,password_input]]):
            new_login_output = self.handler.authenticate_new(
                username = username_input,
                password= password_input
            )
            if not type(new_login_output)==str:
                session_data['current_user'] = new_login_output
                current_user = f"Welcome, {new_login_output['login']}"
                session_data = json.dumps(session_data)
                login_error_div = []
            else:
                session_data = no_update
                current_user = no_update
                login_error_div = dbc.Alert(f'Error logging in with username: {username_input}',color = 'danger')
        else:
            session_data = no_update
            current_user = no_update
            login_error_div = []
        
        return [username_error_div], [password_error_div], [login_error_div], [current_user], session_data

    def submit_create_account(self, clicked,firstname_input, lastname_input, email_input, username_input, password_input,session_data):
        
        
        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        session_data = json.loads(session_data)
        
        firstname_input = get_pattern_matching_value(firstname_input)
        lastname_input = get_pattern_matching_value(lastname_input)
        username_input = get_pattern_matching_value(username_input)
        password_input = get_pattern_matching_value(password_input)
        email_input = get_pattern_matching_value(email_input)

        if username_input is None or username_input=='':
            username_error_div = dbc.Alert('Make sure to enter a username!',color = 'danger')
        else:
            username_error_div = []
        
        if password_input is None or password_input == '':
            password_error_div = dbc.Alert('Make sure to enter your password!',color = 'danger')
        else:
            password_error_div = []

        if email_input is None or email_input == '':
            email_error_div = dbc.Alert('Make sure to enter a valid email address! (And not the same as any other account)',color = 'danger')
        else:
            email_error_div = []

        if not any([i is None or i =='' for i in [firstname_input,lastname_input,email_input,username_input,password_input]]):
            create_user_output = self.handler.create_new_user(
                username = username_input,
                password = password_input,
                email = email_input,
                firstName = firstname_input,
                lastName= lastname_input
            )
            if create_user_output:
                session_data['current_user'] = create_user_output
                current_user = f"Welcome, {create_user_output['login']}"
                session_data = json.dumps(session_data)
                create_account_error_div = []
            else:
                session_data = no_update
                current_user = no_update
                create_account_error_div = dbc.Alert(f'Error creating account with username: {username_input}',color = 'danger')
        else:
            session_data = no_update
            current_user = no_update
            create_account_error_div = []
        
        return [username_error_div],[password_error_div],[email_error_div], [create_account_error_div], [current_user],session_data
        























"""DatasetBuilder Component
"""
import requests
import json
import numpy as np
import pandas as pd

import girder_client

from typing_extensions import Union

from skimage.draw import polygon
from PIL import Image
from io import BytesIO
import base64

# Dash imports
import dash
dash._dash_renderer._set_react_version('18.2.0')
from dash import dcc, callback, ctx, ALL, MATCH, exceptions, Patch, no_update, dash_table
import dash_bootstrap_components as dbc
import dash_mantine_components as dmc
from dash_extensions.enrich import DashBlueprint, html, Input, Output, State, PrefixIdTransform, MultiplexerTransform

from fusion_tools.visualization.vis_utils import get_pattern_matching_value

from fusion_tools import DSATool



class DatasetBuilder(DSATool):
    """Handler for DatasetBuilder component, enabling selection/deselection of folders and slides to add to current visualization session.

    :param DSATool: Sub-class of Tool specific to DSA components. Updates with session data by default.
    :type DSATool: None
    """
    def __init__(self,
                 handler,
                 include_only: Union[list,None] = None
                ):
        
        super().__init__()
        self.include_only = include_only
        self.handler = handler

    def __str__(self):
        return 'Dataset Builder'

    def load(self, component_prefix:int):

        self.component_prefix = component_prefix

        self.title = 'Dataset Builder'
        self.blueprint = DashBlueprint(
            transforms=[
                PrefixIdTransform(prefix=f'{component_prefix}',escape = lambda input_id: self.prefix_escape(input_id)),
                MultiplexerTransform()
            ]
        )

        self.get_callbacks()

        # This might receive the DSAResourceSelector
       
    def gen_collections_dataframe(self,session_data:dict):

        collections_info = []
        collections = self.handler.get_collections()
        for c in collections:
            #slide_count = self.handler.get_collection_slide_count(collection_name = c['name'])
            folder_count = self.handler.get_path_info(path = f'/collection/{c["name"]}')
            #if slide_count>0:
            collections_info.append({
                'Collection Name': c['name'],
                'Collection ID': c['_id'],
                #'Number of Slides': slide_count,
                'Number of Folders': folder_count['nFolders'],
                'Last Updated': folder_count['updated']
            } | c['meta'])

        if 'current_user' in session_data:
            collections_info.append({
                'Collection Name': f'User: {session_data["current_user"]["login"]}',
                'Collection ID': session_data["current_user"]['_id'],
                'Number of Folder': 2,
                'Last Updated': '-',
                'token': session_data['current_user']['token']
            })
            
        collections_df = pd.DataFrame.from_records(collections_info)

        return collections_df

    def update_layout(self, session_data:dict, use_prefix: bool):
        """Generating DatasetBuilder layout

        :return: Div object containing interactive components for the SlideMap object.
        :rtype: dash.html.Div.Div
        """

        # Adding current session data here
        starting_slides = []
        starting_slides_components = []
        starting_slide_idx = 0
        for s in session_data['current']:
            if 'api_url' in s:
                if s['api_url']==self.handler.girderApiUrl:
                    # Getting the id of the DSA slide from this same instance
                    slide_id = s['tiles_url'].split('/item/')[1].split('/')[0]
                    starting_slides.append(slide_id)
                    starting_slides_components.append(
                        self.make_selected_slide(
                            slide_id = slide_id,
                            idx = starting_slide_idx,
                            use_prefix= not use_prefix
                        )
                    )
                    starting_slide_idx += 1
            else:
                thumbnail_address = s['regions_url'].replace('region','thumbnail')
                slide_index = thumbnail_address.split('/')[-3]
                starting_slides.append(f'local{slide_index}')
                starting_slides_components.append(
                    self.make_selected_slide(
                        slide_id = thumbnail_address,
                        idx = starting_slide_idx,
                        local_slide=True,
                        use_prefix = not use_prefix
                    )
                )
                starting_slide_idx+=1

        collections_df = self.gen_collections_dataframe(session_data)

        layout = html.Div([
            dbc.Card(
                dbc.CardBody([
                    dbc.Row(
                        html.H3('Dataset Builder')
                    ),
                    html.Hr(),
                    dbc.Row(
                        'Search through available collections, folders, and slides to assemble a visualization session.'
                    ),
                    html.Hr(),
                    dbc.Row([
                        dcc.Upload(
                            html.Div([
                                'If you have a previous Visualization Session, drag it here or ',
                                html.A('select the file')
                                ]),
                            id = {'type': 'dataset-builder-upload-session','index': 0},
                            accept = 'application/json',
                            style = {
                                'width': '100%',
                                'height': '60px',
                                'lineHeight': '60px',
                                'borderWidth': '1px',
                                'borderStyle': 'dashed',
                                'borderRadius': '5px',
                                'textAlign': 'center',
                                'margin': '10px'
                            },
                            multiple=False
                        )
                    ]),
                    html.Div(
                        dcc.Store(
                            id = {'type':'dataset-builder-data-store','index': 0},
                            storage_type='memory',
                            data = json.dumps({
                                'selected_slides':starting_slides, 
                                'selected_collections': [], 
                                'available_collections': collections_df.to_dict("records")})
                        )
                    ),
                    dbc.Row([
                        html.Div(
                            id = {'type':'dataset-builder-local-slides-div','index': 0},
                            children = [
                                self.make_selectable_dash_table(
                                    dataframe=pd.DataFrame.from_records([
                                        {
                                            'Slide Name': j['name'],
                                            'Slide ID': f'local{j_idx}'
                                        }
                                        for j_idx,j in enumerate(session_data['local'])
                                    ]),
                                    id = {'type':'dataset-builder-slide-table','index': 999},
                                    multi_row= True,
                                    selected_rows = [idx for idx,i in enumerate(session_data['local']) if i in session_data['current']]
                                )
                            ],
                            style = {'marginBottom':'5px'}
                        )
                    ]),
                    html.Hr(),
                    dbc.Row([
                        html.Div(
                            id = {'type': 'dataset-builder-collection-div','index': 0},
                            children = [
                                self.make_selectable_dash_table(
                                    dataframe = collections_df,
                                    id = {'type': 'dataset-builder-collections-table','index': 0},
                                    multi_row = True
                                )
                            ]
                        )
                    ],style={'marginBottom':'10px'}),
                    html.Hr(),
                    dbc.Row([
                        html.Div(
                            id = {'type': 'dataset-builder-collection-contents-div','index':0},
                            children = []
                        )
                    ],style = {'marginBottom':'10px','maxHeight':'70vh','overflow':'scroll'}),
                    html.Div([
                        dmc.Affix([
                            dmc.Accordion([
                                dmc.AccordionItem([
                                    dmc.AccordionControl(
                                        dbc.Stack([
                                            html.I(className='fa-solid fa-microscope',style={'marginRight':'2px'}),
                                            html.H5(f'Visualization Session: ({len(starting_slides)})',style={'textTransform':'none'},id = {'type': 'dataset-builder-vis-session-count','index': 0})
                                        ],direction='horizontal')
                                    ),
                                    dmc.AccordionPanel([
                                        html.Div(
                                            id = {'type':'dataset-builder-selected-slides','index': 0},
                                            children = starting_slides_components,
                                            style = {'maxHeight':'50vh','overflow':'scroll'}
                                        )
                                    ])
                                ],value = 'dataset-builder-vis-session')
                            ],style={'width':'25vw'},radius='lg',variant='separated',chevronPosition='left')
                        ],
                        position = {'bottom':'20','right':'20'})
                    ])
                ])
            )
        ], style = {'maxHeight': '90vh','overflow': 'scroll'})

        if use_prefix:
            PrefixIdTransform(prefix = self.component_prefix).transform_layout(layout)
        
        return layout

    def gen_layout(self, session_data: Union[dict,None]):
        """Generating DatasetBuilder layout, adding to DashBlueprint() object to be embedded in larger layout.

        :param session_data: Data on current session, not used in this component.
        :type session_data: Union[dict,None]
        """

        self.blueprint.layout = self.update_layout(session_data,use_prefix=False)

    def get_callbacks(self):

        # Callback for collection selection (populating table with collection contents)
        self.blueprint.callback(
            [
                Input({'type':'dataset-builder-collections-table','index': ALL},'selected_rows')
            ],
            [
                State({'type':'dataset-builder-data-store','index': ALL},'data'),
                State({'type':'dataset-builder-collection-contents-div','index': ALL},'children'),
                State('anchor-vis-store','data')
            ],
            [
                Output({'type':'dataset-builder-collection-contents-div','index': ALL},'children'),
                Output({'type':'dataset-builder-data-store','index': ALL},'data')
            ],
            prevent_initial_call = True
        )(self.collection_selection)

        # Callback for selecting folder from collection-contents-div
        self.blueprint.callback(
            [
                Input({'type':'dataset-builder-collection-folder-table','index':MATCH},'selected_rows'),
                Input({'type':'dataset-builder-collection-folder-crumb','index':ALL},'n_clicks')
            ],
            [
                State({'type':'dataset-builder-collection-folder-table','index': MATCH},'data'),
                State({'type': 'dataset-builder-collection-folder-nav-parent','index': MATCH},'children'),
                State({'type':'dataset-builder-data-store','index': ALL},'data')
            ],
            [
                Output({'type':'dataset-builder-collection-folder-div','index': MATCH},'style'),
                Output({'type':'dataset-builder-collection-folder-div','index':MATCH},'children'),
                Output({'type':'dataset-builder-collection-slide-div','index':MATCH},'children'),
                Output({'type':'dataset-builder-collection-folder-nav-parent','index': MATCH},'children')
            ],
            prevent_initial_call = True
        )(self.update_folder_div)

        # Callback for selecting slide(s) to be added to visualization session
        self.blueprint.callback(
            [
                Input({'type':'dataset-builder-slide-table','index':ALL},'selected_rows'),
                Input({'type':'dataset-builder-slide-select-all','index':ALL},'n_clicks'),
                Input({'type':'dataset-builder-slide-remove-all','index':ALL},'n_clicks'),
                Input({'type':'dataset-builder-slide-remove-icon','index': ALL},'n_clicks'),
                Input({'type': 'dataset-builder-upload-session','index': ALL},'contents')
            ],
            [
                State({'type': 'dataset-builder-collection-folder-nav-parent','index': ALL},'children'),
                State({'type':'dataset-builder-slide-table','index': ALL},'data'),
                State({'type':'dataset-builder-data-store','index': ALL},'data'),
                State({'type':'dataset-builder-selected-slides','index': ALL},'children'),
                State({'type':'dataset-builder-collection-contents-div','index': ALL},'children'),
                State('anchor-vis-store','data')
            ],
            [
                Output({'type': 'dataset-builder-selected-slides','index': ALL},'children'),
                Output({'type': 'dataset-builder-vis-session-count','index': ALL},'children'),
                Output({'type':'dataset-builder-data-store','index':ALL},'data')
            ]
        )(self.slide_selection)

        # Passing current visualization data to the visualization session for usage in other pages
        self.blueprint.callback(
            [
                Input({'type':'dataset-builder-data-store','index': ALL},'data')
            ],
            [
                State('anchor-vis-store','data')
            ],
            [
                Output('anchor-vis-store','data')
            ]
        )(self.update_vis_store)

        # Callback for plotting slide-level metadata if there is any

    def make_selectable_dash_table(self, dataframe:pd.DataFrame, id:dict, multi_row:bool = True, selected_rows: list = []):
        """Generate a selectable DataTable to add to the layout

        :param dataframe: Pandas DataFrame containing columns/rows of interest
        :type dataframe: pd.DataFrame
        :param id: Dictionary containing "type" and "index" keys for interactivity
        :type id: dict
        :param multi_row: Whether to allow selection of multiple rows in the table or just single, defaults to True
        :type multi_row: bool, optional
        :return: dash_table.DataTable component to be added to layout
        :rtype: dash_table.DataTable
        """

        #Optional: Can hide "ID" columns by adding any column containing "ID" to a list of "hidden_columns"

        dataframe = pd.json_normalize(dataframe.to_dict('records'))
        selectable_table = dash_table.DataTable(
            id = id,
            columns = [{'name':i,'id':i,'deletable':False} for i in dataframe.columns if not i=='token'],
            data = dataframe.to_dict('records'),
            editable = False,
            filter_action='native',
            sort_action = 'native',
            sort_mode = 'multi',
            column_selectable = 'single',
            row_selectable = 'multi' if multi_row else 'single',
            row_deletable = False,
            selected_rows = selected_rows,
            page_action='native',
            page_current=0,
            page_size=10,
            style_cell = {
                'overflow':'hidden',
                'textOverflow':'ellipsis',
                'maxWidth':0                
            },
            tooltip_data = [
                {
                    column: {'value':str(value),'type':'markdown'}
                    for column, value in row.items()
                } for row in dataframe.to_dict('records')
            ],
            tooltip_duration = None
        )

        return selectable_table

    def organize_folder_contents(self, folder_info:dict, show_empty:bool=True, ignore_histoqc:bool=True,session_data:dict = {})->list:
        """For a given folder selection, return a list of slides(0th) and folders (1th)

        :param folder_info: Folder info dict returned by self.handler.get_path_info(path)
        :type folder_info: dict
        :param show_empty: Whether or not to display folders which contain 0 slides, defaults to False
        :type show_empty: bool, optional
        :param session_data: Current session information
        :type session_data: dict
        :return: List of slides within the current folder as well as folders within that folder
        :rtype: list
        """

        folder_folders = []
        folder_slides = []

        # Starting with slides (which will report parent folderId but not that parent's folderId (if applicable))
        if folder_info['_modelType'] in ['folder','collection']:
            all_folder_slides = self.handler.get_folder_slides(
                folder_path = folder_info['_id'],
                folder_type = folder_info['_modelType'],
                ignore_histoqc=ignore_histoqc,
                user_token = session_data['current_user']['token'] if 'current_user' in session_data else None
            )

            folder_slides_folders = [i['folderId'] for i in all_folder_slides]
            unique_folders = list(set(folder_slides_folders))
            folders_in_folder = []
            for u in unique_folders:
                if not u==folder_info['_id'] and not u in folders_in_folder:
                    # This is for all folders in this folder
                    # This grabs parent folders of this folder
                    u_folder_info = self.handler.get_folder_info(
                        folder_id=u,
                        user_token = session_data['current_user']['token'] if 'current_user' in session_data else None
                    )
                    u_folder_rootpath = self.handler.get_folder_rootpath(
                        u,
                        user_token = session_data['current_user']['token'] if 'current_user' in session_data else None
                    )
                    # Folders in order from collection-->child folder-->etc.
                    folder_ids = [i['object']['_id'] for i in u_folder_rootpath]

                    if any([i['object']['_modelType']=='collection' for i in u_folder_rootpath]):
                        base_model = 'collection'
                    else:
                        base_model = 'user'

                    if folder_ids[-1]==folder_info['_id']:
                        child_folder_path = f'/{base_model}/'+'/'.join([i['object']['name'] if i['object']['_modelType'] in ['folder','collection'] else i['object']['login'] for i in u_folder_rootpath]+[u_folder_info['name']])
                    else:
                        # Folder that is immediate child of current folder:
                        child_folder_idx = folder_ids.index(folder_info['_id'])
                        child_folder_path = f'/{base_model}/'+'/'.join([i['object']['name'] if i['object']['_modelType'] in ['folder','collection'] else i['object']['login'] for i in u_folder_rootpath[:child_folder_idx+2]])


                    child_folder_path_info = self.handler.get_path_info(
                        path = child_folder_path,
                        user_token = session_data['current_user']['token'] if 'current_user' in session_data else None
                    )
                    if not child_folder_path_info['_id'] in folders_in_folder:
                        folders_in_folder.append(child_folder_path_info['_id'])
                        
                        # Adding folder to list if the number of items is above zero or show_empty is True
                        folder_folders.append({
                            'Folder Name': child_folder_path_info['name'],
                            'Folder ID': child_folder_path_info['_id'],
                            'Number of Folders': child_folder_path_info['nFolders'],
                            'Number of Slides': child_folder_path_info['nItems'],
                            'Last Updated': child_folder_path_info['updated']
                        } | child_folder_path_info['meta'])

                elif u==folder_info['_id']:
                    # This means that there are some slides that are direct children (not in a sub-folder) in this folder. 
                    # This adds them all at once
                    for i in all_folder_slides:
                        if i['folderId']==folder_info['_id']:
                            folder_slides.append(
                                {
                                    'Slide Name': i['name'],
                                    'Slide ID': i['_id'],
                                    'Last Updated':i['updated']
                                } | {k:v for k,v in i['meta'].items() if type(v)==str}
                            )

        else:
            
            folders_in_folder = []
            unique_folders = []
            user_folders = ['Private','Public']
            for u_f in user_folders:
                user_folder_info = self.handler.get_path_info(
                    path = f'/user/{folder_info["login"]}/{u_f}',
                    user_token = session_data['current_user']['token'] if 'current_user' in session_data else None
                )

                folder_folders.append({
                    'Folder Name': user_folder_info['name'],
                    'Folder ID': user_folder_info['_id'],
                    'Number of Folders': user_folder_info['nFolders'],
                    'Number of Slides': user_folder_info['nItems'],
                    'Last Updated': user_folder_info['updated']
                })

                unique_folders.append(user_folder_info['_id'])


        if show_empty:
            # This is how you get all the empty folders within a folder (does not get child empty folders)
            empty_folders = self.handler.get_folder_folders(
                folder_id = folder_info['_id'],
                folder_type = folder_info['_modelType'],
                user_token = session_data['current_user']['token'] if 'current_user' in session_data else None
            )
            
            for f in empty_folders:
                if not f['_id'] in folders_in_folder and not f['_id'] in unique_folders:
                    if not 'current_user' in session_data:
                        folder_info = self.handler.gc.get(f'/folder/{f["_id"]}/details')
                    else:
                        folder_info = self.handler.gc.get(f'/folder/{f["_id"]}/details?token={session_data["current_user"]["token"]}')
                    folder_folders.append(
                        {
                            'Folder Name': f['name'],
                            'Folder ID': f['_id'],
                            'Number of Folders': folder_info['nFolders'],
                            'Number of Slides': folder_info['nItems'],
                            'Last Updated': f['updated']
                        }
                    )


        return folder_slides, folder_folders

    def make_selected_slide(self, slide_id:str,idx:int,local_slide:bool = False, use_prefix:bool = True):
        """Creating a visualization session component for a selected slide

        :param slide_id: Girder Id for the slide to be added
        :type slide_id: str
        :param local_slide: Whether or not the slide is from the LocalTileServer or if it's in the cloud
        :type local_slide: bool
        :param use_prefix: Whether or not to add the component prefix (initially don't add, when updating the layout do add)
        :param use_prefix: bool
        """
        
        #TODO: Getting individual thumbnails is slow, see if this can be switched to async
        if not local_slide:
            try:
                item_info = self.handler.gc.get(f'/item/{slide_id}')
                item_thumbnail = self.handler.get_image_thumbnail(slide_id)
                folder_info = self.handler.get_folder_info(item_info['folderId'])
                slide_info = {
                    k:v for k,v in item_info.items() if type(v) in [int,float,str]
                }
            except girder_client.HttpError:
                print(f'Item not found! {slide_id}')
                return html.Div()
        else:
            # For local slides, "slide_id" is the request for getting the slide thumbnail
            item_idx = int(slide_id.split('/')[-3])

            try:
                item_thumbnail = Image.open(BytesIO(requests.get(slide_id).content))
                local_names = requests.get(slide_id.replace(f'{item_idx}/tiles/thumbnail','names')).json()['message']
            except (requests.exceptions.ConnectionError, requests.exceptions.RetryError):
                # Triggered on initialization of application because the LocalTileServer instance is not running yet
                item_thumbnail = np.zeros((256,256,3)).astype(np.uint8)
                local_names = ['LOADING LOCALTILESERVER']*(item_idx+1)

            folder_info = {'name': 'Local Slides'}
            slide_info = {'name': local_names[item_idx]}               


        slide_card = html.Div([
            dbc.Card([
                dbc.CardHeader(f"{folder_info['name']}/{slide_info['name']}"),
                dbc.CardBody([
                    dbc.Stack([
                        html.Img(
                            src=Image.fromarray(item_thumbnail) if type(item_thumbnail)==np.ndarray else item_thumbnail
                        ),
                        html.A(
                            html.I(
                                id = {'type': f'{self.component_prefix}-dataset-builder-slide-remove-icon','index': idx} if use_prefix else {'type': 'dataset-builder-slide-remove-icon','index': idx},
                                n_clicks = 0,
                                className = 'bi bi-x-circle-fill fa-2x',
                                style = {'color': 'rgb(255,0,0)','marginRight':'2px'}
                            )
                        )
                    ],direction='horizontal',gap=3)
                ])
            ])
        ],style = {'marginBottom': '2px','width':'25vw'})
        
        return slide_card        

    def collection_selection(self, collection_rows, builder_data, collection_div_children,session_data):
        """Callback for when one/multiple collections are selected from the collections table

        :param collection_rows: Row indices of selected collections
        :type collection_rows: list
        :param builder_data: Data store on available collections and currently included slides
        :type builder_data: list
        :param collection_div_children: Child cards created by collection_selection
        :type collection_div_children: list
        :param session_data: Current session information
        :type session_data: dict
        :return: Children of collection-contents-div (items/folders within selected collections)
        :rtype: list
        """
        selected_collections = get_pattern_matching_value(collection_rows)
        builder_data = json.loads(get_pattern_matching_value(builder_data))
        session_data = json.loads(session_data)

        collection_card_indices = self.get_component_indices(collection_div_children)

        if selected_collections is None and len(builder_data['selected_collections'])==0:
            return ['Select a Collection to get started'], no_update
        elif selected_collections is None and len(builder_data['selected_collections'])>0:
            selected_collections = []

        if len(selected_collections)==1 and len(builder_data['selected_collections'])==0:
            collection_contents = []
        else:
            collection_contents = Patch()

        def add_collection_card(collection_info,idx):
            # For each collection, grab all items and unique folders (as well as those that are not nested in a folder)
            if not 'User: ' in collection_info["Collection Name"]:
                folder_slides, folder_folders = self.organize_folder_contents(
                    folder_info = self.handler.get_path_info(
                        f'/collection/{collection_info["Collection Name"]}',
                        user_token = session_data['current_user']['token'] if 'current_user' in session_data else None
                    ),
                    session_data = session_data
                )
            else:
                folder_slides, folder_folders = self.organize_folder_contents(
                    folder_info=self.handler.get_path_info(
                        f'/user/{collection_info["Collection Name"].replace("User: ","")}',
                        user_token = session_data['current_user']['token'] if 'current_user' in session_data else None
                    ),
                    session_data = session_data
                )

            if len(folder_slides)>0:

                # Checking if any of the slides are already present in the "selected_slides"
                current_selected_slides = builder_data['selected_slides']
                selected_rows = [
                    idx for idx,i in enumerate(folder_slides)
                    if i['Slide ID'] in current_selected_slides
                ]

                non_nested_df = pd.DataFrame.from_records(folder_slides)
                non_nested_table = self.make_selectable_dash_table(
                    dataframe=non_nested_df,
                    id = {'type': f'{self.component_prefix}-dataset-builder-collection-slide-table','index': idx},
                    multi_row=True,
                    selected_rows=selected_rows
                )

            else:
                non_nested_table= html.Div()

            if len(folder_folders)>0:
                folder_table = self.make_selectable_dash_table(
                    dataframe=pd.DataFrame.from_records(folder_folders),
                    id = {'type': f'{self.component_prefix}-dataset-builder-collection-folder-table','index': idx},
                    multi_row=False
                )
            else:
                folder_table = html.Div()

            new_card = html.Div(
                dbc.Card([
                    dbc.CardHeader(f'Collection: {collection_info["Collection Name"]}'),
                    dbc.CardBody([
                        html.H6(
                            children = [
                                dbc.Stack([
                                    html.A('/collection' if not 'User: ' in collection_info['Collection Name'] else '/user'),
                                    html.A(
                                        f'/{collection_info["Collection Name"].replace("User: ","")}/',
                                        id = {'type': f'{self.component_prefix}-dataset-builder-collection-folder-crumb','index': idx},
                                        style = {'color': 'rgb(0,0,255)'}
                                    )
                                ],direction='horizontal')
                            ],
                            id = {'type': f'{self.component_prefix}-dataset-builder-collection-folder-nav-parent','index': idx},
                            style = {'textTransform':'none','display':'inline'}
                        ),
                        html.Hr(),
                        html.Div([
                            folder_table
                        ], id = {'type': f'{self.component_prefix}-dataset-builder-collection-folder-div','index': idx}),
                        html.Hr(),
                        html.Div([
                            non_nested_table
                        ], style = {'marginTop':'10px'},id = {'type': f'{self.component_prefix}-dataset-builder-collection-slide-div','index': idx}),
                        html.Div(
                            dbc.Stack([
                                dbc.Button(
                                    'Select All!',
                                    className = 'd-grid col-6 mx-auto',
                                    n_clicks = 0,
                                    color = 'success',
                                    id = {'type':f'{self.component_prefix}-dataset-builder-slide-select-all','index': idx}
                                ),
                                dbc.Button(
                                    'Remove All',
                                    className = 'd-grid col-6 mx-auto',
                                    n_clicks = 0,
                                    color = 'danger',
                                    id = {'type': f'{self.component_prefix}-dataset-builder-slide-remove-all','index':idx}
                                )
                            ],direction='horizontal',style = {'marginTop':'10px'})
                            if len(folder_slides)>0 else []
                        )
                    ]),
                ]),
                style = {'marginBottom':'10px'},
                id = {'type': f'{self.component_prefix}-dataset-builder-collection-card-div','index': idx}
            )
            
            return new_card
        
        if len(selected_collections)>0:
            if len(list(set(selected_collections).difference(builder_data['selected_collections'])))>0:
                # Adding a new collection
                new_collection_idx = list(set(selected_collections).difference(builder_data['selected_collections']))[0]
                new_component_idx = max(collection_card_indices)+1 if len(collection_card_indices)>0 else 0

                collection_contents.append(add_collection_card(builder_data['available_collections'][new_collection_idx],new_component_idx))
            elif len(list(set(builder_data['selected_collections']).difference(selected_collections)))>0:
                # Removing a collection
                rem_collection_idx = list(set(builder_data['selected_collections']).difference(selected_collections))[0]
                del collection_contents[builder_data['selected_collections'].index(rem_collection_idx)]

        else:
            collection_contents = ['Select a Collection to get started!']

        builder_data['selected_collections'] = selected_collections
        builder_data = json.dumps(builder_data)
            
        return [collection_contents], [builder_data]

    def extract_path_parts(self, current_parts:Union[list,dict], search_key: list = ['props','children'])->tuple:
        """Recursively extract pieces of folder paths stored as clickable components.

        :param current_parts: list or dictionary containing html.A or dbc.Stack of html.A components.
        :type current_parts: Union[list,dict]
        :param search_key: Property keys to search for in nested dicts, defaults to ['props','children']
        :type search_key: list, optional
        :return: Tuple containing all the parts of the folder path
        :rtype: tuple
        """
        path_pieces = ()
        if type(current_parts)==list:
            for c in current_parts:
                if type(c)==dict:
                    for key,value in c.items():
                        if key in search_key:
                            if type(value)==str:
                                path_pieces += (value,)
                            elif type(value) in [list,dict]:
                                path_pieces += self.extract_path_parts(value)
                elif type(c)==list:
                    path_pieces += self.extract_path_parts(c)
        elif type(current_parts)==dict:
            for key,value in current_parts.items():
                if key in search_key:
                    if type(value)==str:
                        path_pieces += (value,)
                    elif type(value) in [list,dict]:
                        path_pieces += self.extract_path_parts(value)

        return path_pieces

    def get_clicked_part(self, current_parts: Union[list,dict])->list:
        """Get the "n_clicks" value for components which have "id". If they have "id" but not "n_clicks", assign 0

        :param current_parts: Either a list or dictionary containing components
        :type current_parts: Union[list,dict]
        :return: List of values corresponding to "n_clicks" 
        :rtype: list
        """
        
        n_clicks_list = []
        if type(current_parts)==list:
            for c in current_parts:
                if type(c)==dict:
                    if 'id' in list(c.keys()):
                        if 'n_clicks' in list(c.keys()):
                            n_clicks_list.append(c['n_clicks'])
                        else:
                            n_clicks_list.append(0)
                    else:
                        for key,value in c.items():
                            if key=='props':
                                n_clicks_list += self.get_clicked_part(value)
                            elif key=='n_clicks':
                                n_clicks_list.append(value)
                elif type(c)==list:
                    n_clicks_list += self.get_clicked_part(c)
        elif type(current_parts)==dict:
            if 'id' in list(current_parts.keys()):
                if 'n_clicks' in list(current_parts.keys()):
                    n_clicks_list.append(current_parts['n_clicks'])
                else:
                    n_clicks_list.append(0)
            else:
                for key,value in current_parts.items():
                    if type(value)==dict:
                        if key=='props':
                            n_clicks_list += self.get_clicked_part(value)
                        elif key=='n_clicks':
                            n_clicks_list.append(value)
                    elif type(value)==list:
                        n_clicks_list += self.get_clicked_part(value)

        return n_clicks_list

    def get_component_indices(self, components: Union[list,dict])->list:
        
        index_list = []
        if type(components)==list:
            for c in components:
                if type(c)==dict:
                    for key,value in c.items():
                        if key=='id':
                            index_list.append(value['index'])
                        elif key in ['props','children']:
                            index_list += self.get_component_indices(value)
                elif type(c)==list:
                    index_list += self.get_component_indices(c)

        elif type(components)==dict:
            for key,value in components.items():
                if key=='id':
                    index_list.append(value['index'])
                elif key in ['props','children']:
                    index_list += self.get_component_indices(value)

        return index_list

    def update_folder_div(self,folder_row,crumb_click,collection_folders,current_crumbs,builder_data):
        """Selecting a folder from the collection's folder table

        :param folder_row: Selected folder (list of 1 index)
        :type folder_row: list
        :param crumb_click: If one of the folder path parts was clicked it will trigger this.
        :type crumb_click: int
        :param collection_folders: Current data in the collection's folder table
        :type collection_folders: list
        :param current_crumbs: List of current path parts that can be selected to go up a folder
        :type current_crumbs: list
        :param builder_data: Current contents of data store for dataset-builder, used for determining if a slide is already selected
        :type builder_data: list
        :return: Sub-folder and slide selection tables for further selection
        :rtype: tuple
        """

        path_parts = self.extract_path_parts(current_crumbs)
        new_crumbs = []
        for i in path_parts:
            if not i in ['/collection','/user']:
                new_crumbs.append(
                    html.A(
                        i,
                        id = {'type': f'{self.component_prefix}-dataset-builder-collection-folder-crumb','index': ctx.triggered_id['index']},
                        style = {'color': 'rgb(0,0,255)'}
                    )
                )
            else:
                new_crumbs.append(
                    html.A(i)
                )

        builder_data = json.loads(get_pattern_matching_value(builder_data))
        
        if 'dataset-builder-collection-folder-table' in ctx.triggered_id['type']:
            
            # Triggers callback when creating new folder table
            if len(folder_row)==0:
                return no_update, no_update, no_update

            new_folder_name = collection_folders[folder_row[0]]['Folder Name']
            folder_table_style = {'display':'inline-block','width':'100%'}
            
            new_crumbs += [
                html.A(
                    new_folder_name + '/',
                    id = {'type': f'{self.component_prefix}-dataset-builder-collection-folder-crumb','index': ctx.triggered_id['index']},
                    style = {'color': 'rgb(0,0,255)'}
                )
            ]
            
            folder_path = ''.join(list(path_parts+(new_folder_name,)))
            folder_info = self.handler.get_path_info(
                path = folder_path
            )
            
            folder_slides, folder_folders = self.organize_folder_contents(
                folder_info=folder_info
            )

            if len(folder_folders)>0:
                folder_table = self.make_selectable_dash_table(
                    dataframe = pd.DataFrame.from_records(folder_folders),
                    id = {'type': f'{self.component_prefix}-dataset-builder-collection-folder-table','index': ctx.triggered_id['index']},
                    multi_row=False
                )
            else:
                folder_table_style = {'display':'none'}
                folder_table = no_update

                
            if len(folder_slides)>0:
                current_selected_slides = builder_data['selected_slides']
                selected_rows = [
                    idx for idx,i in enumerate(folder_slides)
                    if i['Slide ID'] in current_selected_slides
                ]

                slides_table = html.Div([
                    self.make_selectable_dash_table(
                        dataframe=pd.DataFrame.from_records(folder_slides),
                        id = {'type': f'{self.component_prefix}-dataset-builder-slide-table','index': ctx.triggered_id['index']},
                        multi_row=True,
                        selected_rows= selected_rows
                    ),
                    dbc.Stack([
                        dbc.Button(
                            'Select All!',
                            className = 'd-grid col-6 mx-auto',
                            n_clicks = 0,
                            color = 'success',
                            id = {'type':f'{self.component_prefix}-dataset-builder-slide-select-all','index': ctx.triggered_id['index']}
                        ),
                        dbc.Button(
                            'Remove All',
                            className = 'd-grid col-6 mx-auto',
                            n_clicks = 0,
                            color = 'danger',
                            id = {'type': f'{self.component_prefix}-dataset-builder-slide-remove-all','index':ctx.triggered_id['index']}
                        )
                    ],direction='horizontal',style = {'marginTop':'10px'})
                ])
            else:
                slides_table = html.Div()


        elif 'dataset-builder-collection-folder-crumb' in ctx.triggered_id['type']:
            n_clicks = [0]+self.get_clicked_part(current_crumbs)
            n_click_idx = np.argmax(n_clicks)
            if n_click_idx==len(current_crumbs)-1:
                folder_table_style = no_update
                folder_table = no_update
            else:
                folder_table_style = {'display':'inline-block','width':'100%'}
                new_crumbs = new_crumbs[:n_click_idx+1]
                
                new_path = ''.join(list(path_parts)[:n_click_idx+1])[:-1]
                folder_info = self.handler.get_path_info(
                    path = new_path
                )
                
                folder_slides, folder_folders = self.organize_folder_contents(
                    folder_info=folder_info
                )

                if len(folder_folders)>0:
                    folder_table = self.make_selectable_dash_table(
                        dataframe = pd.DataFrame.from_records(folder_folders),
                        id = {'type': f'{self.component_prefix}-dataset-builder-collection-folder-table','index': ctx.triggered_id['index']},
                        multi_row=False
                    )
                else:
                    folder_table = no_update
                    folder_table_style = {'display':'none'}

                    
                if len(folder_slides)>0:

                    current_selected_slides = builder_data['selected_slides']
                    selected_rows = [
                        idx for idx,i in enumerate(folder_slides)
                        if i['Slide ID'] in current_selected_slides
                    ]

                    slides_table = html.Div([
                        self.make_selectable_dash_table(
                            dataframe=pd.DataFrame.from_records(folder_slides),
                            id = {'type': f'{self.component_prefix}-dataset-builder-slide-table','index': ctx.triggered_id['index']},
                            multi_row=True,
                            selected_rows=selected_rows
                        ),
                        dbc.Stack([
                            dbc.Button(
                                'Select All!',
                                className = 'd-grid col-6 mx-auto',
                                n_clicks = 0,
                                color = 'success',
                                id = {'type':f'{self.component_prefix}-dataset-builder-slide-select-all','index': ctx.triggered_id['index']}
                            ),
                            dbc.Button(
                                'Remove All',
                                className = 'd-grid col-6 mx-auto',
                                n_clicks = 0,
                                color = 'danger',
                                id = {'type': f'{self.component_prefix}-dataset-builder-slide-remove-all','index':ctx.triggered_id['index']}
                            )
                        ],direction='horizontal',style = {'marginTop':'10px'})
                    ])
                else:
                    slides_table = html.Div()
                
        new_crumbs = dbc.Stack(new_crumbs,direction='horizontal')

        return folder_table_style, folder_table, slides_table, new_crumbs

    def slide_selection(self, slide_rows, slide_all, slide_rem_all, slide_rem, upload_session, current_crumbs, slide_table_data, builder_data, current_slide_components, current_collection_components, vis_session_data):

        builder_data = json.loads(get_pattern_matching_value(builder_data))
        vis_session_data = json.loads(vis_session_data)

        current_slide_indices = self.get_component_indices(current_slide_components)
        current_collection_indices = list(set(self.get_component_indices(current_collection_components)))

        # When slide-table is not in layout
        if not ctx.triggered_id:
            raise exceptions.PreventUpdate
        
        active_folders = []
        for p in current_crumbs:
            path = ''.join(list(self.extract_path_parts(p)))
            active_folders.append(path)

        current_selected_slides = builder_data['selected_slides']

        selected_slides = Patch()
        
        if 'dataset-builder-slide-table' in ctx.triggered_id['type']:
            # This part has to get triggered for both selection and de-selection of rows in a slide-table
            table_selected_slides = []
            not_selected_slides = []
            for s_r,slide_table in zip(slide_rows,slide_table_data):
                if not s_r is None:
                    table_selected_slides.extend([slide_table[i] for i in s_r])
                    not_selected_slides.extend([slide_table[i] for i in range(len(slide_table)) if not i in s_r])

            new_slides = list(set([i['Slide ID'] for i in table_selected_slides]).difference(current_selected_slides))
            
            for s_idx,s in enumerate(new_slides):
                if not 'local' in s:
                    new_slide_component = self.make_selected_slide(
                        slide_id = s,
                        idx = max(current_slide_indices)+s_idx+1 if len(current_slide_indices)>0 else s_idx
                    )
                else:
                    local_idx = int(s.split('local')[-1])
                    new_slide_component = self.make_selected_slide(
                        slide_id = vis_session_data['local'][local_idx]['regions_url'].replace('region','thumbnail'),
                        idx = max(current_slide_indices)+s_idx+1 if len(current_slide_indices)>0 else s_idx,
                        local_slide=True
                    )

                selected_slides.append(new_slide_component)

            current_selected_slides.extend(new_slides)

            new_rem_slides = list(set(current_selected_slides) & set([i['Slide ID'] for i in not_selected_slides]))
            for d_idx,d in enumerate(new_rem_slides):
                del selected_slides[current_selected_slides.index(d)]
                del current_selected_slides[current_selected_slides.index(d)]           

        elif 'dataset-builder-slide-select-all' in ctx.triggered_id['type']:
            # This part only gets triggered when n_clicks is greater than 0 (ignore trigger on creation)
            if any([i['value'] for i in ctx.triggered]):
                select_all_idx = ctx.triggered_id['index']
                select_all_slides = slide_table_data[current_collection_indices.index(select_all_idx)]

                new_slides = list(set([i['Slide ID'] for i in select_all_slides]).difference(current_selected_slides))
                selected_slides.extend([
                    self.make_selected_slide(
                        slide_id = s,
                        idx = max(current_slide_indices)+s_idx+1 if len(current_slide_indices)>0 else s_idx
                    )
                    for s_idx,s in enumerate(new_slides)
                ])
                current_selected_slides.extend(new_slides)

            else:
                raise exceptions.PreventUpdate

        elif 'dataset-builder-slide-remove-all' in ctx.triggered_id['type']:
            # This part only gets triggered when n_clicks is greater than 0 (ignore trigger on creation)
            if any([i['value'] for i in ctx.triggered]):
                remove_all_idx = ctx.triggered_id['index']
                remove_all_slides = slide_table_data[current_collection_indices.index(remove_all_idx)]

                new_rem_slides = list(set(current_selected_slides) & set([i['Slide ID'] for i in remove_all_slides]))
                for d_idx,d in enumerate(new_rem_slides):
                    del selected_slides[current_selected_slides.index(d)]
                    del current_selected_slides[current_selected_slides.index(d)]

            else:
                raise exceptions.PreventUpdate      
            
        elif 'dataset-builder-slide-remove-icon' in ctx.triggered_id['type']:
            rem_idx = current_slide_indices.index(ctx.triggered_id['index'])
            del selected_slides[rem_idx]
            del current_selected_slides[rem_idx]

        elif 'dataset-builder-upload-session' in ctx.triggered_id['type']:
            upload_session = get_pattern_matching_value(upload_session)
            content_type, content_string = upload_session.split(',')
            try:
                decoded = json.loads(base64.b64decode(content_string))  
            except:
                raise exceptions.PreventUpdate

            if not 'current' in decoded:
                raise exceptions.PreventUpdate

            new_cloud_slides = []
            new_local_slides = []
            keep_slides = []
            for s_idx,s in enumerate(decoded['current']):
                if 'api_url' in s:
                    if s['api_url']==self.handler.girderApiUrl:
                        # Getting the id of the DSA slide from this same instance
                        slide_id = s['tiles_url'].split('/item/')[1].split('/')[0]

                        if not slide_id in current_selected_slides:
                            new_cloud_slides.append(slide_id)
                        else:
                            keep_slides.append(slide_id)
                else:
                    if s['regions_url'] in [i['regions_url'] for i in vis_session_data['local']]:
                        thumbnail_address = s['regions_url'].replace('region','thumbnail')
                        slide_index = thumbnail_address.split('/')[-3]
                        local_slide_id = f'local{slide_index}'
                        if not local_slide_id in current_selected_slides:
                            new_local_slides.append(local_slide_id)
                        else:
                            keep_slides.append(local_slide_id)
            
            # Now removing unneeded slides
            remove_slides = [i for i in current_selected_slides if not i in new_local_slides+new_cloud_slides+keep_slides]
            for d_idx, d in enumerate(remove_slides):
                del selected_slides[current_selected_slides.index(d)]
                del current_selected_slides[current_selected_slides.index(d)]

            for new_idx, new_slide in enumerate(new_local_slides+new_cloud_slides):
                current_selected_slides.append(new_slide)
                if new_slide in new_local_slides:
                    thumbnail_address = vis_session_data['local'][int(new_slide.replace('local',''))]['regions_url'].replace('region','thumbnail')
                    selected_slides.append(
                        self.make_selected_slide(
                            slide_id = thumbnail_address,
                            idx = new_idx,
                            local_slide = True
                        )
                    )
                else:
                    selected_slides.append(
                        self.make_selected_slide(
                            slide_id = new_slide,
                            idx = new_idx
                        )
                    )
            
        else:
            raise exceptions.PreventUpdate
        
        # Need folder id, rootpath, slide info, 
        builder_data['selected_slides'] = current_selected_slides
        builder_data = json.dumps(builder_data)

        # Updated count of included slides:
        included_slide_count = f'Visualization Session ({len(current_selected_slides)})'

        return [selected_slides], [included_slide_count], [builder_data]

    def update_vis_store(self, new_slide_data, current_vis_data):
        """Updating current visualization session based on selected slide(s)

        :param new_slide_data: New slides to be added to the Visualization Session
        :type new_slide_data: list
        :param current_vis_data: Current Visualization Session data
        :type current_vis_data: list
        :return: Updated Visualization Session 
        :rtype: str
        """
        
        new_slide_data = get_pattern_matching_value(new_slide_data)
        if new_slide_data is None:
            raise exceptions.PreventUpdate
        
        new_slide_data = json.loads(get_pattern_matching_value(new_slide_data))
        current_vis_data = json.loads(current_vis_data)
        prev_vis_data_in_handler = []
        for i in current_vis_data['current']:
            if 'api_url' in i:
                if i['api_url']==self.handler.girderApiUrl:
                    prev_vis_data_in_handler.append(i)
            else:
                prev_vis_data_in_handler.append(i)

        # Adding new slides to current_vis_data
        new_slide_info = []
        for s in new_slide_data['selected_slides']:
            if not 'local' in s:
                slide_info = self.handler.gc.get(f'/item/{s}')
                new_slide_info.append(
                    {
                        'name': slide_info['name'],
                        'api_url': self.handler.girderApiUrl,
                        'tiles_url': f'{self.handler.girderApiUrl}/item/{s}/tiles/zxy'+'/{z}/{x}/{y}',
                        'regions_url': f'{self.handler.girderApiUrl}/item/{s}/tiles/region',
                        'metadata_url': f'{self.handler.girderApiUrl}/item/{s}/tiles',
                        'annotations_url': f'{self.handler.girderApiUrl}/annotation/item/{s}',
                        'annotations_metadata_url': f'{self.handler.girderApiUrl}/annotation/?itemId={s}',
                        'annotations_region_url': f'{self.handler.girderApiUrl}/annotation/'
                    }
                )
            else:
                local_idx = int(s.split('local')[-1])
                new_slide_info.append(
                    current_vis_data['local'][local_idx]
                )
        
        prev_slides_in_handler = [i['tiles_url'] for i in prev_vis_data_in_handler]
        new_slides_in_handler = [i['tiles_url'] for i in new_slide_info]

        if sorted(prev_slides_in_handler)==sorted(new_slides_in_handler):
            return no_update
        else:
            new_slides = [new_slide_info[idx] for idx,i in enumerate(new_slides_in_handler) if not i in prev_slides_in_handler]
            rem_slides = [prev_slides_in_handler[idx] for idx,i in enumerate(prev_slides_in_handler) if not i in new_slides_in_handler]

            for s_idx,s in enumerate(current_vis_data['current']):
                if s['tiles_url'] in rem_slides:
                    del current_vis_data['current'][s_idx]
            
            current_vis_data['current'].extend(new_slides)

            return json.dumps(current_vis_data)






